; boot.asm
;
; Code to load F from a disk.

	struc bootblock
.jump		resb 2
.heads		resb 1
.tph		resw 1
.spt		resb 1
.blocks		resd 1
.base		resd 1
.kernel		resd 1
.kernelsize	resb 1
.code		resb 1005
	endstruc

[bits 16]
[section .text]

org 0x7c00
	jmp	short eoh		; jump to end of header

header:
.heads:	db	0x02			; heads
.tph:	dw	0x50			; tracks per head
.spt:	db	0x12			; sectors per track
.blocks:
	dd	0x000005a0		; number of blocks
.base:	dd	0x00000000		; base block
.kernel:
	dd	0x00000001		; kernel block
.kernelsize:
	db	0x01			; kernel size

eoh:
	xor	ax,ax			; let's zero the segments
	mov	ds,ax
	mov	ss,ax
	mov	sp,0xfffe		; temporary stack
	jmp	0x0000:start
start:
	mov	ax,3
	int	0x10			; set video mode 3 (80x25 text)
	call	print
	db	"Loading F...",0x0d,0x0a,0x0d,0x0a,0

	mov	[disk],dl		; save the disk we are booting from

;	mov	si,0x7e00
;	mov	ah,0xc7
;	int	0x15			; read memory info...

	xor	bx,bx
	mov	ax,0x1000
	mov	es,ax			; we're reading to 0x1000:0x0000
	mov	ax,[header.kernel]
	mov	dx,[header.kernel+2]	; dx:ax = block
	shld	dx,ax,1			; multiply dx:ax by 2 to get sector
	shl	ax,1
	mov	di,[header.kernelsize]
	and	di,0x00ff		; di = number of blocks to load
	shl	di,1			; convert to number of sectors
	call	read_sectors		; read kernel

	call	print
	db	"Kernel loaded.",0x0d,0x0a,0

	push	ds
	pop	es			; set es to this segment (0)
	mov	bx,0x8000		; right after this block

	mov	ax,[header.kernel]
	mov	dx,[header.kernel+2]	; dx:ax = start of kernel
	sub	ax,1
	sbb	dx,0			; dx:ax-- (now root block)
	shld	dx,ax,1			; multiply dx:ax by 2 to get sector
	shl	ax,1
	mov	di,2			; just read one block
	call	read_sectors

	mov	bx,0x8000		; we don't need the old data here
	mov	ax,[bx]
	mov	dx,[bx+2]		; dx:ax = first sub-block
	shld	dx,ax,1			; multiply dx:ax by 2 to get sector
	shl	ax,1
	mov	di,2			; read one block
	call	read_sectors

	mov	si,0x8000
	mov	ax,0x8000
	mov	es,ax			; load to 0x8000:0x0000
	xor	bx,bx			
read_block:
	mov	ax,[si]
	mov	dx,[si+2]		; dx:ax = sub-block to load
	cmp	dx,0xffff		; dx = 0xffff?
	jz	.done			; assume ax is 0xffff too and quit :)
	shld	dx,ax,1			; multiply dx:ax by 2 to get sector
	shl	ax,1
	mov	di,2			; read 2 sectors
	call	read_sectors
	add	si,0x10			; go to next header
	jmp	short read_block
.done:
	call	print
	db	"Core blocks loaded.",0x0d,0x0a,0

	cli				; we don't want any interrupts!
	cld				; clear direction flag
	lgdt	[gdt_table]		; load global descriptor table

	mov	eax,cr0
	or	al,1
	mov	cr0,eax			; switch to 32-bit mode

	db	0xea
	dw	start32,8		; jump to 32-bit code

[bits 32]

start32:
	mov	ax,0x10			; data selector
	mov	ds,ax
	mov	es,ax
	mov	ss,ax
	mov	fs,ax
	mov	gs,ax
	mov	esp,0x9f800		; temporary stack at 640kB

	mov	dx,0x3f2		; uhm.. I don't remember where I
	mov	al,0x0c			; stole this from...
	out	dx,al			; wonder what it does :)

	call	a20enable		; DO NOT FORGET@$#!

	mov	esi,0x10000		; kernel loaded at 64kB
	mov	edi,0x4000
	movzx	ecx,byte[0x7c00+bootblock.kernelsize]
	shl	ecx,8			; 256 dwords/block
	rep	movsd			; copy kernel

	mov	eax,0x4000
	jmp	eax

a20enable:
	mov	al,0xd1
	out	0x64,al
	call	a20wait
	mov	al,0xdf
	out	0x60,al
	call	a20wait
	mov	al,0xff
	out	0x64,al
;	call	a20wait
;	ret

a20wait:
	in	al,0x64
	jmp	.l0
.l0:	and	al,2
	jnz	a20wait
	ret

gdt_table:
	dw	gdt_size
	dd	dummy
	dw	0

dummy:	dw	0,0,0,0
code32:
	dw	0xffff,0x0000
	db	0x00,0x9a,0xcf,0x00
data32:
	dw	0xffff,0x0000
	db	0x00,0x92,0xcf,0x00

gdt_size equ $-dummy-1

[bits 16]

print:	pop	si
.again:	lodsb
	mov	ah,0x0e
	or	al,al
	jz	.done
	int	0x10
	jmp	short .again
.done:	jmp	si

; dx:ax = sector, di = count, es:bx = buffer
read_sectors:
	push	di
	push	ax
	push	dx
	call	long_to_bios		; convert to head/track/sector
	mov	dl,[disk]		; dl = disk to load from
.again:	mov	ax,0x0201		; read one sector
	int	0x13
	jc	.again			; on error, try again
	pop	dx
	pop	ax
	add	ax,1			; dx:ax++
	adc	dx,0
	add	bx,0x0200		; move pointer forward by one sector
	pop	di

	dec	di
	jnz	read_sectors		; repeat until di = 0
	ret

; dx:ax = sector
long_to_bios:
	push	ax
	push	dx
	mov	al,[header.heads]	; ax = heads
	xor	ah,ah
	mov	dl,[header.spt]		; dx = sectors per track
	xor	dh,dh
	mul	dx			; ax = heads * spt
	xchg	ax,cx			; cx = ax
	pop	dx
	pop	ax

	div	cx			; divide sector nr by heads*spt
	push	ax			; save track nr
	xor	ax,ax
	xchg	ax,dx			; dx:ax = sector/head
	mov	cl,[header.spt]		; cx = sectors per track
	xor	ch,ch
	div	cx			; dx = sector, ax = head
	xchg	ax,dx
	inc	ax			; sector++
	pop	cx			; cx = track nr

	cmp	byte[disk],0x80		; if HD, disk >= 0x80, else FD
	jb	.fd
	shl	cx,6			; high 10 bits = track
	or	cl,al			; low 6 bits = sector
	jmp	short .hd		; done
.fd:	mov	ch,cl			; high 8 bits = track
	mov	cl,al			; low 8 bits = sector
.hd:	mov	dh,dl			; dh = head
	ret

disk	db	0

times 0x400-($-$$) db 0

