\ This is the part of the F kernel written in Forth. Please note that all
\ numbers are hexadecimal. Comments are mainly for personal use, I refer to
\ a non-existing documentation for further questions.

: immediate ( -- )
  last 4+ 4+ dup c@                     \ get flag byte
  1 or swap c! ;                        \ set immediate bit

: const ( n -- )
  read word header  ['] doconst ,call   \ create word calling doconst
  , immediate ;                         \ compile const, make word immediate

-1 const -1
0 const 0
1 const 1
2 const 2
0 const mode-interpret
1 const mode-compile
2 const mode-postpone

-1 const true
0  const false


: if ( -- a )
  compile ?0branch here 0 , ; immediate \ compile ?0branch to dummy location
: ifnot ( -- a )
  compile ?branch here 0 , ; immediate  \ compile ?branch to dummy location
: else ( a -- a )
  compile branch  here 0 ,              \ compile branch to dummy location
  swap here swap ! ; immediate          \ patch ?branch call of "if"
: then ( a -- )
  here swap ! ; immediate               \ patch ?branch call of "if"/"else"

: var ( n -- )
  read word header  ['] dovar ,call     \ create word calling dovar
  , ;                                   \ compile value
: -> ( n -- )
  mode mode-interpret = if              \ are we in interpretation mode?
    ' >body ! exit                      \ yes, store the word
  then
  mode mode-compile = if                \ are we in compilation mode?
    compile do->  ' , exit              \ yes, compile do-> and ptr to var
  then drop ; immediate
: +-> ( n -- )
  mode mode-interpret = if              \ are we in interpretation mode?
    ' >body +! exit                     \ yes, add to the word
  then
  mode mode-compile = if                \ are we in compilation mode?
    compile do+->  ' , exit             \ yes, compile do+-> and ptr to var
  then drop ; immediate

: char ( -- n )
  read word 1+ c@ ;                     \ read one character and return it
: [char] ( -- )
  compile dolit                         \ read one character and compile it
  char , ; immediate                    \ as a literal

: allot ( n -- )  +-> dp ;              \ add n to dp
: ,"  here 22 parse  here c@ 1+ allot ;

: for  compile >r here ; immediate      \ store count on return stack
: next  compile donext , ; immediate    \ compile a donext to last "for"

: begin  here ; immediate               \ just push return pointer
: again  compile branch , ; immediate   \ jump back to "begin"
: until  compile ?0branch , ; immediate \ jump back to "begin" if false
: repeat  compile ?branch , ; immediate \ jump back to "begin" if true

: kb  0a << ;
: mb  14 << ;

: printcol ( s len -- )
  over c@ 2dup > if  - swap print spaces exit  then  2drop ;
: type ( s count -- )
  for  dup c@ emit  1+ next  drop ;
: emits ( c count -- )
  for  dup emit  next  drop ;
: space-line  50 spaces ;

: dodict ( dict -- )
  read word swap lookup                 \ try to find word
  ?dup ifnot                            \ on error:
    ." Can not find word '" word print
    ." ' in user dictionary" cr         \ print error message
    exit                                \ and return
  then
  mode mode-interpret = if              \ if we're in interpretation mode..
    >code execute exit                  \ then interpret!
  then
  mode mode-compile = if                \ if we're in compilation mode..
    >code , exit                        \ then compile!
  then
  compile docompile , ;                 \ otherwise, postpone compilation
: dictionary ( -- )
  here dup 40 0 fill                    \ allocate 64 words and fill with 0
  100 allot
  create , immediate does> @ dodict ;   \ create word to look up word in dict
: in-use? ( a -- true|false )
  cptr cstack-depth for
    2dup @ = if
      r>drop 2drop  true exit
    then
  4+ next  2drop false ;
: use ( -- )
  ' >body @                             \ get dictionary pointer
  dup in-use? ifnot
    -4 +-> cptr  cptr ! exit            \ push to context stack
  then  drop ;
: ;use ( -- )
  4 +-> cptr ;                          \ drop top of context stack


dictionary fdc
use fdc

1 var sec
0 var cyl
0 var head

: dma ( mode -- )
  cli  06 0a p!  00 0c p!  0b p!        \ initialize DMA, send mode
  00 04 p!  80 04 p!  09 81 p!          \ send offset and page
  00 05 p!  48 05 p!                    \ send length of data
  02 0a p!  sti ;
: status  3f4 p@ ;                      \ get status byte
: send ( byte -- )
  80 begin
    1- ?dup ifnot drop exit then  status c0 and 80 =
  until drop  3f5 p! ;
: recv ( -- byte )
  80 begin
    1- ?dup ifnot 0 exit then  status d0 and d0 =
  until drop  3f5 p@ ;
: ack  recv drop ;                      \ read a return value and drop it
: not-done  0 -> fdc-done ;             \ turns done flag off, wait for int6
: wait ( -- true|false )
  200 for
    fdc-done if r>drop true exit then   \ if an int6 occured, break
    1 ms                                \ 1ms delay
  next false ;                          \ no int6? then break after 500ms
: sense wait drop 08 send ack ack ;     \ sense interrupt cmd
: motor
  fdc-to  ifnot                         \ is the motor off?
    1c 3f2 p!  400 -> fdc-to  200 ms    \ turn it on, wait 500ms
  then  400 -> fdc-to ;                 \ set timeout to 1s
: cal  motor not-done 07 send 00 send sense ; \ recalibrate disk
: seek ( cylinder -- )
  motor not-done ?dup  if               \ cylinder 0?
    0f send 00 send send sense exit     \ no, seek to cylinder
  then  cal  ;                          \ yes, recalibrate
: reset
  0 3f2 p! 200 ms  not-done 0c 3f2 p! sense  0 3f7 p!
  03 send  a1 send  02 send  cal ;
: r/w ( cmd -- )
  motor not-done
  send head 4* send cyl send head send sec send
  02 send sec 1+ send 1b send ff send
  wait ack ack ack ack ack ack ack ;    \ wait for int6, drop return data
: convert ( block -- )
  12 /mod swap -> cyl                   \ divide by blocks per track
  9 /mod 2* 1+ -> sec -> head ;         \ and by blocks per sector
: try-read
  4 for  cyl seek 46 dma e6 r/w  if r>drop exit then  next ;
: read ( address block -- )
  convert  try-read  98000 swap 400 cmove ;
: try-write
  4 for  cyl seek 5a dma c5 r/w  if r>drop exit then  next ;
: write ( address block -- )
  convert
  98000 400 cmove  try-write ;

reset
;use


dictionary block
use block

create bat    400 allot
create buffer 400 allot
false var bat-cached
2 var mainblock

: find ( name mainblock -- ptr block flags true|false )
  buffer swap fdc read                  \ read block to buffer
  buffer begin                          \ look for the block
    dup @ -1 = if                       \ last entry?
      2drop false exit                  \ yes, quit
    then
    2dup 05 + s= if                     \ found it?
      nip dup dup>r @ r> 4+ c@ true exit \ return ptr, block nr and flags
    then
  10 + dup  400 buffer +  = until  drop false ;

: startlist  cr cr ." BLOCKS" cr ." --------------------" cr ;
: endlist  ." --------------------" cr ." END OF LIST" cr ;
: listfile ( ptr -- true|false )
  dup @ dup -1 = if                     \ last block?
    2drop false exit                    \ return false
  then
  swap dup>r 05 + 10 printcol r> 4+ c@ . space . cr true ;
: list ( -- )
  mainblock startlist
  buffer swap fdc read                  \ read block to buffer
  buffer begin
    dup listfile ifnot                  \ if this was the last file, exit
      drop endlist exit
    then
  10 +  dup buffer 400 + =  until       \ repeat until end of block
  drop endlist ;
: findempty ( buffer -- ptr )
  begin
    dup @ -1 = if  exit  then           \ reached the first free entry? exit
  10 + dup  buffer 400 + =  until       \ look until end of block of reached
  drop 0 ;

: read-bat  bat-cached ifnot  bat 1 fdc read  true -> bat-cached  then ;
: write-bat  bat 1 fdc write  false -> bat-cached ;
: free? ( block -- true|false )
  read-bat  bat swap bit 0= ;           \ if bit is set, block is not free
: allocate ( -- block )
  2 begin
    dup free? if  bat over bset  exit  then
  1+ again ;
: free ( block -- )
  read-bat  bat swap bclear ;           \ clear bit in block allocation table
: last-entry ( -- ptr )
  buffer begin
    dup @ -1 = if  exit  then           \ look for the -1 block (last one)
  10 + again ;
: del-entry ( ptr -- )
  last-entry 10 - dup>r swap 10 cmove   \ copy last non-empty block to ptr
  -1 r> ! ;                             \ store -1 to last non-empty block
: free-all ( mainblock -- )
  buffer over fdc read                  \ read the main block
  read-bat  bat swap bclear             \ free the main block
  buffer begin
    dup @ -1 = if
      drop write-bat exit               \ last entry? exit
    then
    dup @ bat swap bclear               \ clear block
    10 +
  dup buffer 400 + =  until  drop ;
: delete ( name mainblock -- true|false )
  dup>r                                 \ mainblock on rstack
  find ifnot  r>drop false exit  then   \ no such name? return false
  1 and if                              \ is this a main block?
    swap del-entry                      \ delete entry in parent block
    buffer r> fdc write                 \ write back to parent block
    free-all true exit                  \ yes, free all sub-blocks
  then
  free del-entry write-bat              \ no main block, free it, del. entry
  buffer r> fdc write                   \ write changes
  true ;

create empty -1 ,
: new ( name flag mainblock -- )
  dup>r buffer swap fdc read            \ read mainblock to buffer
  buffer findempty allocate dup>r over ! \ allocate block and store in entry
  over 1 and if                          \ creating a mainblock?
    empty r> fdc write                  \ yes, write -1 to allocated block
  else
    r>drop
  then
  4+ tuck c!                            \ store flag byte
  -1 over 0c + !                        \ store -1 in next entry
  1+ scopy                              \ store block name
  buffer r> fdc write                   \ write buffer to mainblock
  write-bat ;

: bytes>blocks ( bytes -- blocks )
  dup 0= if
    drop 1 exit
  then  1- 0a >> 1+ ;
: free-rest ( ptr -- )
  begin
    dup @ -1 = if
      drop write-bat exit
    then
    dup @ free
  10 + again ;


\ Note: don't write more than 1kB to a one-block file! If you do, SkyNet
\ will terminate you and your family. Your Computer will rule the world.
\ Powered by Forth. (Yes, I just saw Terminator III).
: wt ( mem size name mainblock -- true|false )
  find ifnot  false exit  then          \ file can't be found, return false
  1 and if                              \ multi-block file?
    nip dup>r buffer swap fdc read      \ read the block
    bytes>blocks buffer swap for        \ write the number of blocks needed
      dup @ -1 = if                     \ do we need to allocate a new block?
        allocate over !                 \ store new block nr
        0 over 4+ !                     \ flags = 0, name length = 0
        -1 over 10 + !                  \ make next entry last one
      then
      2dup @ fdc write                  \ write block
      10 +  swap  400 +  swap           \ next entry and block
    next
    nip free-rest                       \ drop mem, free remaining blocks
    buffer r> fdc write                 \ write back main block
    write-bat exit
  then
  nip nip  fdc write  true ;            \ write this only block

\ You can't just read parts of a file. That's for modern systems,
\ like... CP/M.
: rd ( mem name mainblock -- true|false )
  find ifnot  false exit  then
  1 and if
    nip buffer swap fdc read            \ read main block
    buffer begin
      dup @ -1 = if                     \ last entry?
        drop  true exit                 \ yes, return true
      then
      2dup @ fdc read                   \ read the block
    10 + swap  400 + swap  again
  then
  nip fdc read  true ;

: load ( mem -- true|false )
  read word mainblock rd ;
: save ( mem size -- true|false )
  read word mainblock find ifnot        \ is the file there?
    word 1 mainblock new                \ no, create it
  else
    3drop                               \ yes, go on and write it
  then
  word mainblock wt ;

: root  2 -> mainblock ;

: run ( -- )
  read                                  \ read filename
  10000 malloc fcreate                  \ create mem file
  fptr @ @ word mainblock rd ifnot      \ load data to it
    fpop ffree  ." Can not load file!" cr
  then ;

;use

: dir 2 block list ;

dictionary editor
use editor

\ See the documentation for scancodes.

10000 malloc var buffer

1 var min-y
0 var min-x
15 var max-y
4f var max-x
0 var size-x
0 var size-y
: calc-size ( -- )
  max-x min-x - 1+ -> size-x
  max-y min-y - 1+ -> size-y ;
calc-size

0 var line
min-x var cursor-x
min-y var cursor-y
0 var ascii
0 var scan
0 var size
buffer var pos
false var active
true var screen-dirty
false var line-dirty
false var buf-dirty

0 buffer c!

: asciizlen ( s -- n )
  0 begin
    over c@ 0= if  nip exit  then
    1+ swap  1+ swap
  again ;

: new ( -- )
  0 buffer c!
  buffer -> pos
  0 -> line
  min-x -> cursor-x
  min-y -> cursor-y
  false -> buf-dirty
  true -> screen-dirty
  false -> line-dirty ;
: load ( -- )
  new
  buffer block load
  ifnot
    ." Can not load file." cr exit
  then
  buffer asciizlen -> size ;
: save ( -- )
  buffer size block save
  ifnot
    ." Can not save file." cr exit
  then  false -> buf-dirty ;

: next-line ( a -- a )
  begin
    dup c@ 0a = if
      1+ exit                           \ if new line, return ptr to next char
    then
    dup c@ ifnot
      exit                              \ if zero, return ptr to it
    then
  1+ again ;
: start-line ( a -- a )
  begin
    dup buffer = if                     \ OK, this is ugly...
      exit
    then
    1-                                  \ go to last char
    dup c@ 0a = if                      \ new line?
      1+ exit                           \ yes, return the next char
    then
  again ;
: last-line ( a -- )
  start-line 1- start-line ;
: line>addr ( line -- address )
  ?dup ifnot  buffer exit  then         \ if line = 0, return start of text
  buffer swap for                       \ otherwise, do "line" times:
    next-line dup c@ ifnot              \ go to next line
      r>drop exit                       \ if end of file, return
    then
  next ;
: last-line? ( -- true|false )
  pos begin
    dup c@ ifnot
      drop true exit
    then
    dup c@ 0a = if
      drop false exit
    then
  1+ again ;
: eol? ( -- true|false )
  pos c@ 0= if  true exit  then
  pos c@ 0a = if  true exit  then
  false ;
: linepos ( -- n )
  0 pos begin
    1- dup c@ 0a = if
      drop exit
    then
    dup buffer = if
      drop exit
    then
    swap 1+ swap
  again ;

: insert ( c -- )
  pos pos 1+ size buffer + pos - 1+
  cmove-back                            \ make space
  pos c!                                \ store char
  1 +-> size ;                          \ one more byte
: delete ( -- )
  pos 1+ pos size buffer + pos - 1+ cmove
  -1 +-> size ;

: print-line ( a -- a )
  begin
    dup c@ ifnot  exit  then            \ if zero, exit
    dup c@ emit                         \ print character
    dup c@ 0a = if  1+ exit  then       \ if new line, increase address, exit
  1+ again ;
: status
  min-x max-y 1+ at
  2d 50 emits
  min-x max-y 1+ 1+ 2dup at
  space-line  at
  ." Cursor: " cursor-x min-x - . ." ," line cursor-y + min-y - .
  ."  | Position: " pos buffer - . ." /" size .
  ."  bytes (" size 0a >> . ."  kB)" ;
: print-screen ( -- )
  screen-dirty ifnot
    status
    line-dirty if
      line cursor-y + min-y - line>addr
      min-x cursor-y 2dup at space-line at
      print-line drop
      false -> line-dirty
    then
    exit
  then
  false -> line-dirty
  false -> screen-dirty
  clear min-x min-y at                  \ locate cursor at first line
  line line>addr
  size-y for
    print-line dup c@ ifnot             \ print line, and check for end of file
      r>drop drop status exit           \ if EOF, return
    then
  next drop  status ;
: set-cursor ( -- )
  cursor-x cursor-y at ;
: getkey ( -- )
  key-scan  -> ascii  7f and -> scan ;
: line-up ( -- )
  pos last-line -> pos  min-x -> cursor-x ;
: line-down ( -- )
  pos next-line -> pos  min-x -> cursor-x ;
: do-up ( -- )
  cursor-y min-y = if
    line 0= if
      exit
    then
    -1 +-> line  true -> screen-dirty  line-up  exit
  then
  -1 +-> cursor-y line-up ;
: do-down ( -- )
  last-line? ifnot
    cursor-y max-y = if
      1 +-> line  true -> screen-dirty  line-down  exit
    then
    1 +-> cursor-y line-down
  then ;
: do-left ( -- )
  cursor-x min-x = ifnot
    -1 +-> cursor-x  -1 +-> pos
  then ;
: do-right ( -- )
  eol? ifnot
    cursor-x max-x = ifnot
      1 +-> cursor-x  1 +-> pos
    then 
  then ;
: do-pgup ( -- )
  size-y 1- for  do-up  next ;
: do-pgdown ( -- )
  size-y 1- for  do-down  next ;
: do-home ( -- )
  pos start-line -> pos  min-x -> cursor-x ;
: do-end ( -- )
  pos next-line dup c@ if
    1-
  then -> pos
  linepos min-x + -> cursor-x ;
: do-del ( -- )
  pos buffer - size = ifnot
    pos c@ 0a = if
      delete  true -> screen-dirty  exit
    then
    delete  true -> line-dirty
  then ;
: do-bksp ( -- )
  pos buffer = ifnot
  true -> line-dirty
    -1 +-> pos
    pos c@ 0a = if
      true -> screen-dirty
      linepos -> cursor-x
      cursor-y min-y = if
        line if
          -1 +-> line
        then
      else
        -1 +-> cursor-y
      then
    else
      -1 +-> cursor-x
    then
    delete
  then ;
: do-esc ( -- )
  false -> active ;
: do-enter ( -- )
  0a insert do-down  true -> screen-dirty ;
: do-tab ( -- )
  4 for  20 insert do-right  next  true -> line-dirty ;
: check ( -- )
  scan 48 = if do-up exit then
  scan 50 = if do-down exit then
  scan 4b = if do-left exit then
  scan 4d = if do-right exit then
  scan 49 = if do-pgup exit then
  scan 51 = if do-pgdown exit then
  scan 47 = if do-home exit then
  scan 4f = if do-end exit then
  scan 53 = if do-del exit then
  scan 0e = if do-bksp exit then
  scan 01 = if do-esc exit then
  ascii 0a = if do-enter exit then
  ascii 09 = if do-tab exit then
  ascii if
    ascii insert do-right
    ascii 0a = if
      true -> screen-dirty  exit
    then
    true -> line-dirty
  then ;
: edit ( -- )
  true -> active  true -> screen-dirty
  begin
    print-screen set-cursor getkey check
  active repeat ;

;use

: welcome
  ." The F system core is now compiled and ready for use." cr
  ." Questions or comments? You can reach me at " 0c attrib
  ." robost@home.se" 07 attrib cr
  ." The latest F code and documentation can be found at "
  0c attrib ." http://robos.org/" 07 attrib cr
  cr ." Hint: type "
  0f attrib ." use editor  load core  edit"
  07 attrib ."  to see the system source code. Enjoy!" cr ;

welcome

