; compiler.asm
;
; The Forth compiler.

%define mode_interpret	0
%define mode_compile	1
%define mode_postpone	2

%define flag_immediate	1

	struc memfile
.data	resd	1			; pointer to data
.offset	resd	1			; offset within data block
	endstruc


core_header 3c,_fptr,"fptr",flag_none
_fptr:
	call	_dovar
	dd	fstack+(4*7)

fstack	dd	0,0,0,0,0,0,0,0

; fpush ( memfile -- )
core_header 3d,_fpush,"fpush",flag_none
_fpush:
	sub	dword[_fptr+5],byte 4
	mov	eax,[_fptr+5]
	mov	[eax],ebx
	pop	ebx
	next

; fpop ( -- memfile )
core_header 3c,_fpop,"fpop",flag_none
_fpop:
	mov	eax,[_fptr+5]
	add	dword[_fptr+5],byte 4
	push	ebx
	mov	ebx,[eax]
	next

; ffree ( memfile -- )
core_header 29,_ffree,"ffree",flag_none
_ffree:
	mov	eax,[ebx+memfile.data]
	test	eax,0xfff00000
	jz	.done
	call	mem_free
.done:	mov	eax,ebx
	call	mem_free
	pop	ebx
	next

; fcreate ( data -- )
core_header 25,_fcreate,"fcreate",flag_none
_fcreate:
	push	esi
	mov	eax,memfile_size
	call	mem_alloc
	mov	dword[esi+memfile.data],ebx
	mov	dword[esi+memfile.offset],0
	sub	dword[_fptr+5],byte 4
	mov	eax,[_fptr+5]
	mov	[eax],esi
	pop	esi
	pop	ebx
	next

; fread ( memfile -- c )
core_header 01,_fread,"fread",flag_none
_fread:
	mov	eax,[ebx+memfile.data]
	add	eax,[ebx+memfile.offset]
	inc	dword[ebx+memfile.offset]
	movzx	ebx,byte[eax]
	next

; OK.. even I admit it, this is UGLY! However, it does work :)
; bkey ( -- c )
core_header 22,_bkey,"bkey",flag_none
_bkey:
	call	_nest
.next:	dd	_kbuf,_cfetch		; kbuf c@
	dd	_bptr,_eq		; bptr =
	dd	_cbranch,.eol		; if goto .eol then
	dd	_kbuf,_cfetch,_1add	; kbuf c@ 1+
	dd	_bptr,_eq		; bptr =
	dd	_cbranch,.eol1		; if goto .eol1 then
	dd	_bptr,_1add,_dup	; bptr 1+ dup
	dd	_doto,_bptr		; -> bptr
	dd	_kbuf,_add,_cfetch	; kbuf + c@
	dd	_exit
.eol:	dd	_dolit,1,_doaddto,_bptr	; 1 +-> bptr
	dd	_dolit,0x20,_exit	; 0x20 exit ( return space )
.eol1:	dd	_cr,_kbuf,_dolit,0x7f	; cr kbuf 0x7f
	dd	_accept			; accept
	dd	_dolit,0,_doto,_bptr	; 0 -> bptr
	dd	_dolit,0x0a,_exit	; 0x0a exit ( return new line )

; input ( -- c )
core_header 05,_input,"input",flag_none
_input:
	call	_nest
.read:	dd	_fptr,_dolit,fstack+(8*4) ; fptr fstack+(8*4)
	dd	_eq			; =
	dd	_cbranch,.keyboard	; if goto .keyboard then
	dd	_fptr,_fetch,_fread	; fptr @ fread
	dd	_cdup			; ?dup
	dd	_cbranch,.done		; if goto .done then
	dd	_fptr,_fetch,_ffree	; fptr @ ffree
	dd	_dolit,4
	dd	_doaddto,_fptr		; 4 +-> fptr
	dd	_branch,.read		; goto .read
.keyboard:
	dd	_bkey
.done:	dd	_exit

; parse ( string char -- )
core_header 07,_parse,"parse",flag_none
_parse:
	call	_nest
	dd	_swap,_duptor,_1add	; swap dup>r 1+
.next:	dd	_input,_duptor,_2pick,_eq ; input dup>r 2pick =
	dd	_cbranch,.end		; if goto .end then
	dd	_rto,_over,_cstore,_1add ; r> over c! 1+
	dd	_branch,.next		; goto .next
.end:	dd	_rtodrop		; r>drop
	dd	_rfetch,_sub,_1sub	; r@ - 1-
	dd	_rto,_cstore		; r> c!
	dd	_drop,_exit		; drop exit

; blank? ( c -- true | false )
core_header 26,_blankc,"blank?",flag_none
_blankc:
	cmp	bl,0x20
	jz	.blank
	cmp	bl,0x0a
	jz	.blank
	cmp	bl,0x09
	jz	.blank
	xor	ebx,ebx
	next
.blank:	mov	ebx,-1
	next

; read ( -- )
core_header 16,_read,"read",flag_none
_read:
	call	_nest
.loop0:	dd	_input,_dup,_blankc	; input dup blank?
	dd	_c0branch,.done0	; ifnot goto .done0 then
	dd	_drop,_branch,.loop0	; drop goto .loop0
.done0:	dd	_word,_1add,_swap	; word 1+ swap
.loop1:	dd	_over,_cstore,_1add	; over c! 1+
	dd	_input,_dup,_blankc	; input dup blank?
	dd	_c0branch,.loop1	; ifnot goto .loop1 then
	dd	_drop			; drop
	dd	_word,_1add,_sub	; word 1+ -
	dd	_word,_cstore,_exit	; word c! exit

; lookup ( name dict -- header | false )
core_header 14,_lookup,"lookup",flag_none
_lookup:
	call	_nest

	dd	_over,_hash,_add,_fetch	; over hash + @
.next:	dd	_dup,_c0branch,.error
	dd	_2dup			; 2dup  ( NOS = name searched )	
	dd	_dolit,9,_add		; 9 +   ( TOS = header name )
	dd	_seq			; s=    ( compare strings )
	dd	_cbranch,.found		; jump to .found if equal
	dd	_fetch,_dup		; @ dup ( get next header )
	dd	_cbranch,.next		; if not last header, jump to .next
.error:	dd	_2drop,_dolit,0,_exit	; 2drop 0 exit ( error )
.found:	dd	_nip,_exit		; nip exit     ( success )

; find ( name -- header | false )
core_header 2e,_find,"find",flag_none
_find:
	call	_nest
	dd	_cptr			; cptr
.again:	dd	_dup,_dolit,cstack+(8*4),_eq
	dd	_cbranch,.done		; dup cstack+8*4 = if goto .done then
	dd	_dup,_fetch		; dup @  ( get dictionary header )
	dd	_2pick,_swap		; 2pick swap  ( name cptr name dict )
	dd	_lookup			; lookup ( returns 0 if not found )
	dd	_dup,_cbranch,.found	; dup if goto .found then
	dd	_drop,_4add		; 4+ ( add 4 to cptr )
	dd	_branch,.again		; try this dictionary
.found:
	dd	_2nip,_exit		; 2nip exit ( return header )
.done:
	dd	_2drop,_dolit,0,_exit	; 2drop 0 exit ( return false )

; mode_interpret var mode
core_header 16,_mode,"mode",flag_none
_mode:
	call	_dovar
	dd	mode_interpret

; heap var dp
; ' dp alias here
core_header 2e,_dp,"here",flag_none
core_header 32,_dp,"dp",flag_none
_here:
_dp:
	call	_dovar
	dd	heap

; cstack+7*4 var cptr
core_header 30,_cptr,"cptr",flag_none
_cptr:
	call	_dovar
	dd	cstack + (7*4)

core_header 3e,_cstack_depth,"cstack-depth",flag_none
_cstack_depth:
	push	ebx
	mov	ebx,cstack+(8*4)
	sub	ebx,[_cptr+5]
	shr	ebx,2
	next

; 0 var last
core_header 36,_last,"last",flag_none
_last:
	call	_dovar
	dd	0

; 16 var base
core_header 0e,_base,"base",flag_none
_base:
	call	_dovar
	dd	0x10

core_header 26,_dict,"dict",flag_none
_dict:
	push	ebx
	mov	ebx,[_cptr+5]		; ebx = pointer to dict stack
	mov	ebx,[ebx]		; ebx = dictionary currently used
	next

; dictpos ( name -- ptr to dictionary list )
core_header 29,_dictpos,"dictpos",flag_none
_dictpos:
	call	_nest
	dd	_hash,_dict,_add	; hash dict +
	dd	_exit			; exit

core_header 19,_comma,",",flag_none
_comma:
	mov	eax,[_dp+5]		; eax = dp
	add	dword[_dp+5],byte 4	; dp += 4
	mov	[eax],ebx		; store TOS
	pop	ebx			; pop new TOS
	next

core_header 37,_docompile,"docompile",flag_none
_docompile:
	lodsd				; eax = word to compile
	mov	ecx,[_dp+5]		; ecx = dp
	add	dword[_dp+5],byte 4	; dp += 4
	mov	[ecx],eax		; store word
	next

core_header 26,_ccomma,"c,",flag_none
_ccomma:
	mov	eax,[_dp+5]		; eax = dp
	inc	dword[_dp+5]		; dp++
	mov	[eax],bl		; store TOS
	pop	ebx			; pop new TOS
	next

core_header 26,_scomma,"s,",flag_none
_scomma:
	mov	eax,[_dp+5]		; eax = dp
	movzx	ecx,byte[ebx]		; ecx = string length
	mov	[eax],cl		; store length
.next:	inc	eax			; eax++ (dp)
	inc	ebx			; ebx++ (string to store)
	mov	dl,[ebx]		; al = character to copy
	mov	[eax],dl		; store to heap
	loop	.next
	inc	eax			; eax++ (dp)
	mov	[_dp+5],eax		; store dp
	pop	ebx			; pop new TOS
	next

; header ( name -- )
core_header 30,_header,"header",flag_none
_header:
	call	_nest
	dd	_here,_doto,_last	; here -> last
	dd	_here,_over,_dictpos	; here over dictpos
	dd	_dup,_fetch		; dup @ ( get ptr to last header )
	dd	_comma			; , ( store it on heap )
	dd	_store			; ! ( store old "here" to "dictpos" )
	dd	_dup,_cfetch		; dup c@ ( get length of name )
	dd	_dolit,10,_add,_here,_add ; 10 + dp +
	dd	_comma,_dolit,0,_ccomma	; , 0 c, ( store ptr to code & flags )
	dd	_scomma			; s, ( store name )
	dd	_last,_comma,_exit	; last , exit ( ptr to this header )

; ,call ( a -- )
core_header 3b,_commacall,",call",flag_none
_commacall:
	mov	ecx,[_dp+5]		; ecx = dp
	add	dword[_dp+5],byte 5	; dp += 5
	mov	byte[ecx],0xe8		; compile call instruction
	sub	ebx,ecx
	sub	ebx,byte 5		; ebx = relative address
	mov	[ecx+1],ebx		; compile address
	pop	ebx
	next

; create ( -- )
core_header 36,_create,"create",flag_none
_create:
	call	_nest
	dd	_read,_word,_header	; read word header
	dd	_dolit,_dovariable	; ['] dovariable literal
	dd	_commacall		; ,call
	dd	_exit			; exit

core_header 37,_imode,"[",flag_immediate
_imode:
	mov	byte[_mode+5],mode_interpret
	next

core_header 3b,_cmode,"]",flag_immediate
_cmode:
	mov	byte[_mode+5],mode_compile
	next

core_header 35,_colon,":",flag_none
_colon:
	call	_nest
	dd	_read,_word,_header	; read word header
	dd	_dolit,_nest		; ['] nest literal
	dd	_commacall		; ,call
	dd	_cmode			; ]
	dd	_exit			; exit

core_header 37,_semicolon,";",flag_immediate
_semicolon:
	call	_nest
	dd	_docompile,_exit	; compile exit
	dd	_imode			; [
	dd	_exit			; exit

; core_header 34,_docode,"docode",flag_none
_docode:
	call	_nest
	dd	_rto			; r>
	dd	_last,_4add,_fetch	; last 4+ @
	dd	_1add,_duptor,_4add	; 1+ dup>r 4+
	dd	_sub,_rto,_store	; - r> !
	dd	_exit			; exit

core_header 34,_dodoes,"dodoes",flag_none
_dodoes:
	xchg	ebp,esp
	push	esi			; save IP on return stack
	xchg	ebp,esp
	pop	esi			; get IP from hardware stack
	xchg	ebx,[esp]		; swap
	next

core_header 33,_doesto,"does>",flag_immediate
_doesto:
	call	_nest
	dd	_docompile,_docode	; compile docode
	dd	_dolit,_dodoes		; ['] dodoes
	dd	_commacall		; ,call
	dd	_exit			; exit

core_header 31,_compile,"compile",flag_immediate
_compile:
	call	_nest
	dd	_docompile,_docompile	; [compile] docompile
	dd	_tick,_comma,_exit	; ' , exit

core_header 3b,_compileimm,"[compile]",flag_immediate
_compileimm:
	call	_nest
	dd	_tick,_comma,_exit	; ' , exit

; ' ( -- a )
core_header 0f,_tick,"'",flag_none
_tick:
	call	_nest
	dd	_read,_word,_find	; read word find
	dd	_cdup			; ?dup
	dd	_cbranch,.ok		; if goto .ok then
	dd	_doprint
	db	0x21,"Word could not be found by tick: "
	dd	_cr,_exit		; cr exit
.ok:	dd	_4add,_fetch		; 4+ @
	dd	_exit			; exit

core_header 3d,_tickimm,"[']",flag_immediate
_tickimm:
	call	_nest
	dd	_docompile,_dolit	; [compile] dolit
	dd	_tick,_comma,_exit	; ' , exit

; >header ( code -- header )
core_header 0f,_toheader,">header",flag_none
_toheader:
	mov	ebx,[ebx-4]
	next

; >body ( code -- body )
core_header 01,_tobody,">body",flag_none
_tobody:
	add	ebx,byte 5
	next

; >code ( header -- code )
core_header 03,_tocode,">code",flag_none
_tocode:
	mov	ebx,[ebx+4]
	next

; process ( name -- true | false )
core_header 2b,_process,"process",flag_none
_process:
	call	_nest
	dd	_dup,_find,_cdup	; dup find ?dup
	dd	_cbranch,.isword	; if goto .isword then

	dd	_ston			; s>n
	dd	_cbranch,.number	; if goto .number then

	dd	_doprint		; ." Word not found: "
	db	0x11,0x0a,"Word not found: "
	dd	_word,_print,_cr	; word print cr
	dd	_dolit,0,_exit		; 0 exit
.number:
	dd	_mode,_dolit,mode_interpret
	dd	_eq			; mode mode_interpret =
	dd	_cbranch,.number_interpret ; if goto .number_interpret then
	dd	_mode,_dolit,mode_postpone
	dd	_eq			; mode mode_postpone =
	dd	_cbranch,.number_postpone ; if goto .number_postpone then
.number_compile:
	dd	_docompile,_dolit
	dd	_comma			; [compile] dolit ,
.number_interpret:
	dd	_dolit,-1,_exit		; -1 exit
.number_postpone:
	dd	_docompile,_docompile	; [compile] docompile
	dd	_docompile,_dolit	; [compile] dolit
	dd	_docompile,_docompile	; [compile] docompile
	dd	_comma			; ,
	dd	_dolit,-1,_exit		; -1 exit

.isword:
	dd	_nip,_mode		; nip mode
	dd	_dolit,mode_interpret	; mode_interpret
	dd	_eq			; =
	dd	_cbranch,.interpret	; if goto .interpret then
	dd	_mode			; mode
	dd	_dolit,mode_postpone	; mode_postpone
	dd	_eq			; =
	dd	_cbranch,.postpone	; if goto .postpone then
	dd	_dup,_4add,_4add	; dup 8 +  ( get to flag byte )
	dd	_cfetch			; c@
	dd	_dolit,flag_immediate	; flag_immediate
	dd	_and,_cbranch,.interpret ; and if goto .interpret then
.compile:
	dd	_4add,_fetch,_comma	; 4+ @ ,  ( compile the word )
	dd	_dolit,-1,_exit		; -1 exit ( return true )
.interpret:
	dd	_4add,_fetch		; 4+ @
	dd	_execute		; execute
	dd	_dolit,-1,_exit		; -1 exit ( return true )
.postpone:
	dd	_4add,_fetch		; 4+ @
	dd	_docompile,_docompile	; [compile] docompile
	dd	_comma			; ,	
	dd	_dolit,-1,_exit		; -1 exit ( return true )

core_header 11,_paren,"(",flag_immediate
_paren:
	call	_nest
	dd	_dolit,1		; 1
.next:	dd	_input			; input
	dd	_dup,_dolit,'(',_eq	; dup '(' =
	dd	_cbranch,.levelup	; if goto .levelup then
	dd	_dolit,')',_eq		; ')' =
	dd	_cbranch,.leveldown	; if goto .leveldown then
	dd	_branch,.next		; goto .next
.levelup:
	dd	_drop,_1add		; drop 1+
	dd	_branch,.next		; goto .next
.leveldown:
	dd	_1sub,_cdup		; 1- ?dup
	dd	_cbranch,.next		; if goto .next then
	dd	_exit

core_header 39,_backslash,"\",flag_immediate
_backslash:
	call	_nest
.next:	dd	_input,_dolit,0x0a,_eq	; input 0x0a =
	dd	_c0branch,.next		; ifnot goto .next then
	dd	_exit

core_header 35,_doconst,"doconst",flag_none
_doconst:
	pop	eax
	mov	eax,[eax]		; eax = constant
	mov	ecx,[_dp+5]		; ecx = dp
	cmp	byte[_mode+5],mode_interpret
	jz	.interpret
	cmp	byte[_mode+5],mode_compile
	jz	.compile
.postpone:
	add	dword[_dp+5],byte 0x10	; compile 0x10 bytes
	mov	dword[ecx],_docompile
	mov	dword[ecx+0x04],_dolit	; compile dolit
	mov	dword[ecx+0x08],_docompile
	mov	dword[ecx+0x0c],eax	; compile constant
	next
.interpret:
	push	ebx
	xchg	eax,ebx			; push constant
	next
.compile:
	add	dword[_dp+5],byte 8	; compile 8 bytes
	mov	dword[ecx],_dolit
	mov	dword[ecx+4],eax	; compile dolit constant
	next

core_header 2c,_bptr,"bptr",flag_none
_bptr:
	call	_dovar
	dd	0

core_header 34,_kbuf,"kbuf",flag_none
_kbuf:
	call	_dovariable
	times	0x80 db 0

core_header 3e,_word,"word",flag_none
_word:
	call	_dovariable
	times	0x40 db 0

cstack	dd	0,0,0,0,0,0,0,coredict

