; crt.asm
;
; Text mode routines. Taken from Primula with minor modifications.

	struc screen
.x	resb	1
.y	resb	1
.curx	resb	1
.cury	resb	1
.attrib	resb	1
.active	resb	1
.memory	resd	1
.mode	resd	1
	endstruc

crt_init:
	mov	ebx,main_screen
	call	crt_create_screen
	mov	dword[main_screen+screen.memory],0xb8000
	mov	byte[main_screen+screen.active],1
	call	crt_clear_screen
	ret

; crt_create_screen
;
; create a virtual screen
; remember that screen buffer memory is NOT allocated
;
; ebx = pointer to screen struct
;
; returns:
; all registers unchanged

crt_create_screen:
	push	eax
	mov	eax,dword[crt_mode]
	mov	dword[ebx+screen.mode],eax
	mov	al,byte[crt_x]
	mov	byte[ebx+screen.x],al
	mov	ah,byte[crt_y]
	mov	byte[ebx+screen.y],ah
	mov	byte[ebx+screen.curx],0
	mov	byte[ebx+screen.cury],0
	mov	byte[ebx+screen.attrib],0x07
	mov	byte[ebx+screen.active],0
	pop	eax
	ret

; crt_load_screen
;
; Loads the contents of the virtual screen to the physical.
;
; ebx = pointer to screen struct
;
; returns:
; all registers unchanged

crt_load_screen:
	push	eax
	push	ecx
	push	esi
	push	edi

	mov	edi,0xb8000
	mov	esi,dword[ebx+screen.memory]
	mov	cl,byte[ebx+screen.x]
	mov	al,byte[ebx+screen.y]
	mul	cl
	movzx	ecx,ax
	rep	movsw

	pop	edi
	pop	esi
	pop	ecx
	pop	eax
	ret

; crt_save_screen
;
; Saves the contents of the physical screen to the virtual screen.
;
; ebx = pointer to screen struct
;
; returns:
; all registers unchanged

crt_save_screen:
	push	eax
	push	ecx
	push	esi
	push	edi

	mov	esi,0xb8000
	mov	edi,dword[ebx+screen.memory]
	mov	cl,byte[ebx+screen.x]
	mov	al,byte[ebx+screen.y]
	mul	cl
	movzx	ecx,ax
	rep	movsw

	pop	edi
	pop	esi
	pop	ecx
	pop	eax
	ret

; crt_clear_screen
;
; Clears the screen pointed to by ebx
;
; ebx = pointer to screen struct
;
; returns:
; all registers unchanged

crt_clear_screen:
	push	eax
	push	ecx
	push	edi

	mov	cl,byte[ebx+screen.x]
	mov	al,byte[ebx+screen.y]
	mul	cl
	movzx	ecx,ax			; ecx = (x size * y size)

	call	crt_get_address		; get address to write to

	mov	ah,byte[ebx+screen.attrib]
	mov	al,' '			; al = space
	rep	stosw			; fill it :)

	xor	ecx,ecx
	call	crt_set_cursor

	pop	edi
	pop	ecx
	pop	eax

	ret

; crt_get_address
;
; ebx = pointer to screen struct
;
; returns:
; edi = address of video buffer (or physical is screen.active is set)

crt_get_address:
	cmp	byte[ebx+screen.active],0
	jnz	.active			; is this screen mapped to physical video memory?
	mov	edi,dword[ebx+screen.memory]
	ret				; nope, return the virtual buffer
.active:
	mov	edi,0xb8000		; yeah, use physical video mem
	ret

; crt_set_cursor
;
; set virtual and physical cursor
;
; ebx = pointer to screen struct
; cl = x
; ch = y
;
; returns:
; all registers unchanged

crt_set_cursor:

	mov	byte[ebx+screen.curx],cl
	mov	byte[ebx+screen.cury],ch
	call	crt_set_cursor_physical

	ret

; crt_put_string
;
; print a counted string
;
; ebx = pointer to screen struct
; esi = pointer to string
;
; returns:
; all registers unchanged

crt_put_string:
	push	eax
	push	ecx
	push	edx
	push	esi

	lodsb
	movzx	ecx,al
	or	ecx,ecx
	jz	.quit
.again:
	lodsb
	mov	dl,al
	call	crt_put_char
	loop	.again
.quit:
	pop	esi
	pop	edx
	pop	ecx
	pop	eax
	ret

; crt_put_stringz
;
; print a zero terminated string
;
; ebx = pointer to screen struct
; esi = pointer to string
;
; returns:
; all registers unchanged

crt_put_stringz:
	push	eax
	push	edx
	push	esi
.again:

	lodsb
	or	al,al
	jz	.quit

	mov	dl,al
	call	crt_put_char

	jmp	short .again
.quit:
	pop	esi
	pop	edx
	pop	eax
	ret

; crt_put_char
;
; put a char on the current cursor position
;
; ebx = pointer to screen struct
; dl = char
;
; returns:
; all registers unchanged

crt_put_char:
	push	eax
	push	ecx

	cmp	dl,0x0a
	jnz	.not_0a

	call	crt_move_cursor_down
	mov	byte[ebx+screen.curx],0
	call	crt_set_cursor_physical
	jmp	short .quit
.not_0a:
	mov	al,byte[ebx+screen.x]
	mov	ah,byte[ebx+screen.cury]
	mul	ah
	movzx	eax,ax
	movzx	ecx,byte[ebx+screen.curx]
	add	eax,ecx
	shl	eax,1

	call	crt_get_address

	add	edi,eax
	mov	al,dl
	mov	ah,byte[ebx+screen.attrib]
	mov	word[edi],ax
	call	crt_move_cursor_forward
.quit:
	pop	ecx
	pop	eax
	ret

; crt_set_char_pos
;
; put a char... duh
;
; cl = x
; ch = y
; ebx = pointer to screen struct
; dl = char
;
; returns:
; all registers unchanged

crt_set_char_pos:
	push	eax
	push	ecx
	push	edi

	mov	al,byte[ebx+screen.x]
	mov	ah,ch
	mul	ah
	movzx	eax,ax
	movzx	ecx,cl
	add	eax,ecx
	shl	eax,1

	call	crt_get_address

	add	edi,eax
	mov	al,dl
	mov	ah,byte[ebx+screen.attrib]
	mov	word[edi],ax

	pop	edi
	pop	ecx
	pop	eax
	ret

; crt_get_char_pos
;
; get a char on the screen...
;
; cl = x
; ch = y
; ebx = pointer to screen struct
;
; returns:
; dl = char

crt_get_char_pos:
	push	eax
	push	ecx
	push	edi

	mov	al,byte[ebx+screen.x]
	mov	ah,ch
	mul	ah
	movzx	eax,ax
	movzx	ecx,cl
	add	eax,ecx
	shl	eax,1

	call	crt_get_address

	add	edi,eax
	mov	al,byte[edi]

	pop	edi
	pop	ecx
	pop	eax

	ret

; crt_set_attrib_pos
;
; set the attribute on a certain position of the screen
;
; cl = x
; ch = y
; ebx = pointer to screen struct
; dl = attrib
;
; returns:
; all registers unchanged

crt_set_attrib_pos:
	push	eax
	push	ecx
	push	edi

	mov	al,byte[ebx+screen.x]
	mov	ah,ch
	mul	ah
	movzx	eax,ax
	movzx	ecx,cl
	add	eax,ecx
	shl	eax,1

	call	crt_get_address

	add	edi,eax

	mov	byte[edi+1],dl

	pop	edi
	pop	ecx
	pop	eax
	ret

; crt_get_attrib_pos
;
; set the attribute on a certain position of the screen
;
; cl = x
; ch = y
; ebx = pointer to screen struct
;
; returns:
; dl = attrib

crt_get_attrib_pos:
	push	eax
	push	ecx
	push	edi

	mov	al,byte[ebx+screen.x]
	mov	ah,ch
	mul	ah
	movzx	eax,ax
	movzx	ecx,cl
	add	eax,ecx
	shl	eax,1

	call	crt_get_address

	add	edi,eax

	mov	al,byte[edi+1]

	pop	edi
	pop	ecx
	pop	eax

	ret

; crt_move_cursor_forward
;
; Moves the cursor one step forward, scrolls down if needed.
;
; ebx = pointer to screen struct
;
; returns:
; all registers unchanged

crt_move_cursor_forward:
	push	eax

	mov	al,byte[ebx+screen.curx]
	inc	eax
	cmp	al,byte[ebx+screen.x]
	jb	.ok

	call	crt_move_cursor_down
	xor	al,al
.ok:
	mov	byte[ebx+screen.curx],al
	call	crt_set_cursor_physical

	pop	eax
	ret

; crt_move_cursor_back
;
; Moves the cursor one step back
;
; ebx = pointer to screen struct
;
; returns:
; all registers unchanged

crt_move_cursor_back:
	push	eax

	mov	al,byte[ebx+screen.curx]
	or	al,al
	jz	.done

	dec	eax
	mov	byte[ebx+screen.curx],al
	call	crt_set_cursor_physical
.done:
	pop	eax
	ret

; crt_move_cursor_down
;
; Moves the cursor one row down, scrolls down if needed.
;
; ebx = pointer to screen struct
;
; returns:
; all registers unchanged

crt_move_cursor_down:
	push	eax

	mov	al,byte[ebx+screen.cury]
	inc	eax
	cmp	al,byte[ebx+screen.y]
	jb	.ok

	call	crt_scroll_down
	jmp	short .quit
.ok:
	mov	byte[ebx+screen.cury],al
	call	crt_set_cursor_physical
.quit:
	pop	eax
	ret

; crt_move_cursor_up
;
; Moves the cursor one row up
;
; ebx = pointer to screen struct
;
; returns:
; all registers unchanged

crt_move_cursor_up:
	push	eax

	mov	al,byte[ebx+screen.cury]
	or	al,al
	jz	.done

	dec	eax
	mov	byte[ebx+screen.cury],al
	call	crt_set_cursor_physical
.done:
	pop	eax
	ret

; crt_scroll_down
;
; Scroll down one row :)
;
; ebx = pointer to screen struct
;
; returns:
; all registers unchanged

crt_scroll_down:
	push	eax
	push	ecx
	push	esi
	push	edi

	call	crt_get_address

	mov	al,byte[ebx+screen.x]
	mov	ah,byte[ebx+screen.y]
	dec	ah
	mul	ah
	movzx	ecx,ax
	;shl	ecx,1

	mov	al,byte[ebx+screen.x]
	movzx	eax,al
	shl	eax,1

	mov	esi,edi
	add	esi,eax

	;rep	movsb
	rep	movsw

	mov	cl,byte[ebx+screen.x]
	movzx	ecx,cl
	mov	al,' '
	mov	ah,byte[ebx+screen.attrib]
	rep	stosw

	pop	edi
	pop	esi
	pop	ecx
	pop	eax

	ret

crt_set_cursor_physical:
	push	eax
	push	ecx
	push	edx

	cmp	byte[ebx+screen.active],0
	jz	.quit

	mov	al,byte[ebx+screen.cury]
	mov	ah,byte[ebx+screen.x]
	mul	ah
	movzx	eax,ax

	movzx	ecx,byte[ebx+screen.curx]
	add	eax,ecx
	xchg	eax,ecx

	mov	dx,0x03d4
	mov	al,0x0e
	mov	ah,ch
	out	dx,ax

	inc	eax
	mov	ah,cl
	out	dx,ax
.quit:
	pop	edx
	pop	ecx
	pop	eax
	ret

crt_mode	dd	3
crt_x		db	80
crt_y		db	25

