; crtforth.asm
;
; Some glue to the functions in crt.asm and a bunch of useful I/O words

; clear ( -- )
core_header 29,_clear,"clear",flag_none
_clear:
	push	ebx
	mov	ebx,main_screen
	call	crt_clear_screen
	pop	ebx
	next

; emit ( c -- )
core_header 32,_emit,"emit",flag_none
_emit:
	mov	dl,bl			; dl = character to print
	mov	ebx,main_screen
	call	crt_put_char
	pop	ebx			; pop new TOS
	next

; print ( s -- )
core_header 29,_print,"print",flag_none
_print:
	push	esi			; save IP
	mov	esi,ebx			; esi = string pointer
	mov	ebx,main_screen
	call	crt_put_string
	pop	esi			; restore IP
	pop	ebx			; pop new TOS
	next

; at ( x y -- )
core_header 2e,_at,"at",flag_none
_at:
	pop	ecx			; cl = x
	mov	ch,bl			; ch = y
	mov	ebx,main_screen
	call	crt_set_cursor
	pop	ebx			; pop new TOS
	next

; attrib ( a -- )
core_header 32,_attrib,"attrib",flag_none
_attrib:
	mov	[main_screen+screen.attrib],bl
	pop	ebx
	next

; n>s ( n s -- )
core_header 37,_ntos,"n>s",flag_none
_ntos:
	pop	eax			; eax = n
	push	ebx			; save start of string
	inc	ebx
	movzx	ecx,byte[_base+5]	; ecx = base of number system
.next:	xor	edx,edx			; edx:eax = n
	div	ecx			; eax = new n, edx = remainder
	sub	dl,0x0a			; dl = dl - 10
	jc	.digit			; is it a digit (below 10)?
	add	dl,'a'			; no, it's a "character"
	jmp	short .char
.digit:	add	dl,0x0a+'0'		; it's a digit
.char:	mov	[ebx],dl		; store digit/character
	inc	ebx			; go to next location in buffer
	or	eax,eax			; is n = 0?
	jnz	.next			; nope, repeat

	; OK, now we're done here. There's just a slight problem - the
	; string is stored backwards! It doesn't have any count byte yet,
	; either.

	pop	eax			; eax = start of string
	mov	ecx,ebx
	sub	ecx,eax			; ecx = end of string - start of string
	dec	ecx
	mov	[eax],cl		; store count byte
	inc	eax			; eax = ptr to first char of string
	dec	ebx			; ebx = ptr to last char of string
.swap:	cmp	eax,ebx			; eax >= ebx?
	jnb	.done			; yes, we're done
	mov	cl,[eax]
	xchg	cl,[ebx]
	mov	[eax],cl		; swap characters
	inc	eax
	dec	ebx			; next character pair
	jmp	short .swap
.done:	pop	ebx			; pop new TOS
	next

; . ( n -- )
core_header 1d,_dot,".",flag_none
_dot:
	call	_nest
	dd	_dolit,dot_buffer
	dd	_ntos			; dot_buffer n>s
	dd	_dolit,dot_buffer
	dd	_print,_exit		; dot_buffer print exit

; cr ( -- )
core_header 32,_cr,"cr",flag_none
_cr:
	call	_nest
	dd	_dolit,0x0a,_emit,_exit	; 0x0a emit exit


; spaces ( n -- )
core_header 32,_spaces,"spaces",flag_none
_spaces:
	call	_nest
	dd	_tor
.next:	dd	_dolit,0x20,_emit,_donext,.next
	dd	_exit

; space ( -- )
core_header 31,_space,"space",flag_none
_space:
	call	_nest
	dd	_dolit,1,_spaces,_exit

; doprint ( -- )
core_header 35,_doprint,"doprint",flag_none
_doprint:
	push	ebx			; save TOS
	lodsb
	movzx	ecx,al			; ecx = length of string
	or	ecx,ecx
	jz	.done			; just in case some moron tries...
	mov	ebx,main_screen
.next:	lodsb				; get next character
	mov	dl,al
	call	crt_put_char		; print it
	loop	.next
.done:	pop	ebx			; restore TOS
	next

; ." ( -- )
core_header 3e,_dotq,'."',flag_immediate
_dotq:
	call	_nest
	dd	_docompile,_doprint	; compile doprint
	dd	_here,_dolit,0x22	; here '"'
	dd	_parse			; parse
	dd	_here,_cfetch,_1add	; here c@ 1+
	dd	_doaddto,_dp		; +-> dp
	dd	_exit			; exit

