; f.asm
;
; Entry file for the F Forth environment.

; 2003-10-24:
;
; Happy UN day, Robert! Don't include the "next" macro AFTER using it.
; NASM thinks that means you're defining a label!
;
; 2003-12-16:
;
; If I put the kernel at location 0, it works in bochs but not on a real
; PC. If I locate it at 0x40000, it works on a real computer but not in
; bochs. At 0x800, it crashes both. I hate life.
;
; 2003-12-17 (00:54)
;
; I traced the error back to the "lookup" function in compiler.asm.. I
; simply forgot to check if "dict" is 0. I love life. I hate myself.

[bits 32]
[section .text]
[org 0x4000]

%define memory_size	0x10000		; 1kB blocks > 1MB
%define my_stack_size	0x1000		; in bytes
%define rstack_size	0x400		; in bytes
%define pit_stack_size	0x400		; won't need much more than 1kB

%define registers_size	152

start:
	call	gdt_init
	call	mem_init
	call	idt_init
	call	irq_init
	call	kbd_init
	call	crt_init
	call	pit_init
	call	fdc_init

	mov	eax,my_stack_size
	call	mem_alloc
	mov	esp,esi
	add	esp,my_stack_size
	mov	edx,esi
	call	proc_init		; enter multitasking mode

	mov	eax,memfile_size
	call	mem_alloc

	mov	[fstack+(7*4)],esi
	mov	dword[esi+memfile.data],0x80000
	mov	dword[esi+memfile.offset],0

	mov	eax,rstack_size
	call	mem_alloc

	mov	eax,[proc_current]
	mov	eax,[eax+proc_struct.stack]
	mov	dword[eax+mtforth_struct.code],0
	mov	dword[eax+mtforth_struct.data],0
	mov	dword[eax+mtforth_struct.rstack],esi
	add	esi,rstack_size
	mov	ebp,esi
	mov	dword[eax+mtforth_struct.rsize],rstack_size

	sti				; do not forget :)

mainword:
	call	_nest
	dd	_doprint
	db	0x20,"F kernel started. Compiling...",0xa,0xa
.next:
	dd	_read,_word,_process,_drop
	dd	_branch,.next

%include "coreheader.asm"
%include "core.asm"

%include "gdt.asm"
%include "idt.asm"
%include "irq.asm"

; %include "mm.asm"
%include "newmm.asm"

%include "kbd.asm"
%include "crt.asm"
%include "pit.asm"
%include "timer.asm"
%include "proc.asm"
%include "ipc.asm"
%include "fdc.asm"

%include "dict.asm"
%include "timerforth.asm"
%include "mmforth.asm"
%include "mtforth.asm"
%include "ipcforth.asm"
%include "kbdforth.asm"
%include "compiler.asm"
%include "crtforth.asm"

%include "cpu.asm"
%include "ports.asm"
%include "var.asm"
%include "logic.asm"
%include "jump.asm"
%include "mem.asm"
%include "math.asm"
%include "stack.asm"

; This file has to be included after all word definitions.

%include "coredict.asm"

[section .bss]

proc_registers	resb	registers_size
pit_stack	resb	pit_stack_size
kbd_buffer	resb	kbd_buffer_size
main_screen	resb	screen_size
dot_buffer	resb	40
idt		resb	0x800

heap:

