; kbd.asm
;
; Keyboards can be useful.

; kbd_special   
; 
; bit 0         left shift
; bit 1         right shift
; bit 2         left ctrl
; bit 3         right crtl
; bit 4         alt
; bit 5         alt gr

; kbd_status
;
; bit 0         scroll lock
; bit 1         num lock
; bit 2         caps lock

; keymap entry
;
; 0000  byte    scan code (set bit 7 for prefix)
; 0001  byte    normal ascii
; 0002  byte    shifted ascii
; 0003  byte    alt (gr)ed ascii
; 0004  byte    shifted and alt (gr)ed ascii
; 0005  byte    code with caps lock active (and nothing else)
; 0006  byte    code with num lock active (and nothing else)
; 0007  byte    reserved


%define kbd_buffer_size 0x40

%include "keymap_se.asm"
%include "keymap_de.asm"
%include "keymap_us.asm"

%define kbd_default_map keymap_us

	struc keymap
.scancode	resb	1
.normal		resb	1
.shift		resb	1
.alt		resb	1
.shiftalt	resb	1
.caps		resb	1
.num		resb	1
.reserved	resb	1
	endstruc

kbd_init:
	push	ebx

	mov	byte[kbd_special],0
	mov	byte[kbd_status],0
	mov	byte[kbd_prefix],0
	mov	eax,0x21
	mov	ebx,kbd_int
	call	idt_set_int		; hook interrupt
	mov	cl,1
	call	irq_enable		; enable IRQ1
	call	kbd_update_leds		; update LEDs

	pop	ebx
	ret

kbd_int:
	push	eax
	push	esi

	in	al,0x60
	cmp	al,0xfa
	jnz	.no_ack_here
	mov	byte[kbd_ack],al
	jmp	.no_ack
.no_ack_here:
	mov	byte[kbd_ack],0
	cmp	al,0xe0
	jnz	.no_prefix
	mov	byte[kbd_prefix],al
	jmp	.quit_prefix
.no_prefix:
	cmp	al,0x53
	jnz	.no_reboot2
	push	eax
	mov	al,byte[kbd_special]
	and	al,0x14
	cmp	al,0x14
	jnz	.no_reboot
	call	kbd_wait
	mov	al,0xfe
	out	0x64,al
	jmp	short $-2
.no_reboot:
	pop	eax
.no_reboot2:
	cmp	al,0x2a
	jz	.do_lshift
	cmp	al,0x36
	jz	.do_rshift
	cmp	al,0xaa
	jz	.do_lshiftr
	cmp	al,0xb6
	jz	.do_rshiftr
	cmp	al,0x1d
	jz	.do_ctrl
	cmp	al,0x9d
	jz	.do_ctrlr
	cmp	al,0x38
	jz	.do_alt
	cmp	al,0xb8
	jz	near .do_altr
	cmp	al,0x3a
	jz	near .do_caps
	cmp	al,0x45
	jz	near .do_num
	cmp	al,0x46
	jz	near .do_scroll
	jmp	.normal_key
.do_lshift:
	or	byte[kbd_special],0x01
	jmp	.special_done
.do_rshift:
	or	byte[kbd_special],0x02
	jmp	.special_done
.do_lshiftr:
	and	byte[kbd_special],0xfe
	jmp	short .special_done
.do_rshiftr:
	and	byte[kbd_special],0xfd
	jmp	short .special_done
.do_ctrl:
	cmp	byte[kbd_prefix],0
	jz	.do_lctrl
.do_rctrl:
	or	byte[kbd_special],0x08
	jmp	short .special_done
.do_lctrl:
	or	byte[kbd_special],0x04
	jmp	short .special_done
.do_ctrlr:
	and	byte[kbd_special],0xfc
	jmp	short .special_done
.do_alt:
	cmp	byte[kbd_prefix],0
	jz	.do_lalt
.do_ralt:
	or	byte[kbd_special],0x20
	jmp	short .special_done
.do_lalt:
	or	byte[kbd_special],0x10
	jmp	short .special_done
.do_altr:
	and	byte[kbd_special],0xcf
	jmp	short .special_done
.do_caps:
	xor	byte[kbd_status],0x04
	jmp	short .special_done
.do_num:
	xor	byte[kbd_status],0x02
	jmp	short .special_done
.do_scroll:
	xor	byte[kbd_status],0x01
;	jmp	short .special_done
.special_done:
	jmp	.quit

.normal_key:
	cmp	al,0x80
	jnb	near .quit
	cmp	byte[kbd_prefix],0
	jz	.dont_add_80
	or	al,0x80
.dont_add_80:
	mov	ah,al
	xor	al,al
	mov	esi,dword[kbd_map]
.try_key:
	cmp	byte[esi+keymap.scancode],0
	jz	.just_scan
	cmp	ah,byte[esi+keymap.scancode]
	jnz	.next_key
	test	byte[kbd_special],0x33
	jnz	.ascii_shift_or_alt
	test	byte[kbd_status],0x04
	jnz	.ascii_caps
	test	byte[kbd_status],0x02
	jnz	.ascii_num
.ascii_normal:
	mov	al,byte[esi+keymap.normal]
	jmp	short .key_found
.ascii_caps:
	mov	al,byte[esi+keymap.caps]
	jmp	short .key_found
.ascii_num:
	mov	al,byte[esi+keymap.num]
	jmp	short .key_found
.ascii_shift_or_alt:
	test	byte[kbd_special],0x03
	jz	.ascii_alt
	test	byte[kbd_special],0x30
	jz	.ascii_shift
.ascii_both:
	mov	al,byte[esi+keymap.shiftalt]
	jmp	short .key_found
.ascii_shift:
	mov	al,byte[esi+keymap.shift]
	jmp	short .key_found
.ascii_alt:
	mov	al,byte[esi+keymap.alt]
	jmp	short .key_found
.next_key:
	add	esi,byte keymap_size
	jmp	short .try_key
.key_found:
	or	al,al
	jz	.quit

.just_scan:
	call	kbd_write
.quit:
	mov	byte[kbd_prefix],0
	call	kbd_update_leds
.quit_prefix:
	in	al,0x61
	mov	ah,al
	or	al,0x80
	out	0x61,al
	mov	al,ah
	out	0x61,al			; send ack to kbc
.no_ack:
	mov	al,0x20
	out	0x20,al			; send ack to PIC

	pop	esi
	pop	eax
	iret

; al = ASCII, ah = scancode
kbd_write:
	push	ecx
	mov	ecx,[kbd_inptr]
	mov	[ecx],ax
	add	ecx,byte 2
	cmp	ecx,kbd_buffer+kbd_buffer_size
	jb	.ok
	mov	ecx,kbd_buffer
.ok:	mov	[kbd_inptr],ecx
	pop	ecx
	ret

; returns ASCII in al, scan code in ah
kbd_read:
	push	ecx
.wait:	mov	ecx,[kbd_outptr]
	cmp	ecx,[kbd_inptr]
	jz	.wait
	mov	ax,[ecx]
	add	ecx,byte 2
	cmp	ecx,kbd_buffer+kbd_buffer_size
	jb	.ok
	mov	ecx,kbd_buffer
.ok:	mov	[kbd_outptr],ecx

	movzx	eax,ax
	pop	ecx
	ret

; returns eax = -1 if true, else 0.
kbd_poll:
	mov	eax,[kbd_inptr]
	sub	eax,[kbd_outptr]
	jz	.done
	mov	eax,-1
.done:	ret

kbd_update_leds:
	push	eax
	call	kbd_wait
	mov	al,0xed
	out	0x60,al
	call	kbd_wait
	mov	al,byte[kbd_status]
	and	al,0x07
	out	0x60,al
	pop	eax
	ret

kbd_wait:
	push	eax
.wait:	in	al,0x64
	test	al,0x02
	jnz	.wait
	pop	eax
	ret


core_header 3f,_keymap_us,"keymap-us",flag_none
_keymap_us:
	call	_doconst
	dd	keymap_us

core_header 3f,_keymap_se,"keymap-se",flag_none
_keymap_se:
	call	_doconst
	dd	keymap_se

core_header 3f,_keymap_de,"keymap-de",flag_none
_keymap_de:
	call	_doconst
	dd	keymap_de

; kbd_default_map var keymap
core_header 3c,_keymap,"keymap",flag_none
_keymap:
	call	_dovar
kbd_map		dd	kbd_default_map

; 0 var key-status
core_header 00,_key_status,"key-status",flag_none
_key_status:
	call	_dovar
kbd_status	dd	0

; 0 var key-special
core_header 01,_key_special,"key-special",flag_none
_key_special:
	call	_dovar
kbd_special	dd	0

kbd_prefix	db	0
kbd_ack		db	0

kbd_inptr	dd	kbd_buffer
kbd_outptr	dd	kbd_buffer

