; mem.asm
;
; Fetching, storing and other memory manipulation.

core_header 01,_fetch,"@",flag_none
_fetch:
	mov	ebx,[ebx]
	next

core_header 03,_store,"!",flag_none
_store:
	pop	eax
	mov	[ebx],eax
	pop	ebx
	next

core_header 30,_addstore,"+!",flag_none
_addstore:
	pop	eax
	add	[ebx],eax
	pop	ebx
	next

core_header 0e,_cfetch,"c@",flag_none
_cfetch:
	movzx	ebx,byte[ebx]
	next

core_header 10,_cstore,"c!",flag_none
_cstore:
	pop	eax
	mov	[ebx],al
	pop	ebx
	next

core_header 1e,_wfetch,"w@",flag_none
_wfetch:
	movzx	ebx,word[ebx]
	next

core_header 20,_wstore,"w!",flag_none
_wstore:
	pop	eax
	mov	[ebx],ax
	pop	ebx
	next

; cmove ( source dest length -- )
core_header 2b,_cmove,"cmove",flag_none
_cmove:
	or	ebx,ebx
	jz	.exit
	mov	ecx,ebx			; number of bytes to copy
	mov	eax,esi			; save esi
	mov	edx,edi			; save edi
	pop	edi			; destination address
	pop	esi			; source address
	shr	ecx,2			; convert byte count to dwords
	rep	movsd			; copy the data
	mov	ecx,ebx			; ecx = byte count
	and	ecx,3			; modulo 4
	rep	movsb			; copy remaining bytes, if any
	mov	esi,eax			; restore esi
	mov	edi,edx			; restore edi
	pop	ebx
	next
.exit:
	add	esp,byte 8
	pop	ebx
	next

; cmove-back ( source dest length -- )
core_header 30,_cmove_back,"cmove-back",flag_none
_cmove_back:
	or	ebx,ebx
	jz	_cmove.exit
	std
	mov	ecx,ebx			; number of bytes to copy
	mov	eax,esi			; save esi
	mov	edx,edi			; save edi
	pop	edi			; destination address
	pop	esi			; source address
	add	esi,ecx
	add	edi,ecx
	shr	ecx,2			; convert byte count to dwords
	jz	.no_dword
	sub	esi,byte 4
	sub	edi,byte 4
	rep	movsd			; copy the data
.dword_done:
	mov	ecx,ebx			; ecx = byte count
	and	ecx,3			; modulo 4
	jz	.done
	add	esi,byte 3
	add	edi,byte 3
	jmp	short .do_bytes
.no_dword:
	mov	ecx,ebx
.do_bytes:
	rep	movsb			; copy remaining bytes, if any
.done:
	cld
	mov	esi,eax			; restore esi
	mov	edi,edx			; restore edi
	pop	ebx
	next

; fill ( address count word -- )
core_header 2e,_fill,"fill",flag_none
_fill:
	mov	edx,edi			; save edi
	mov	eax,ebx			; eax = word to fill with
	pop	ecx			; ecx = count
	pop	edi			; edi = address
	jecxz	.done			; if count = 0, skip
	rep	stosd			; store the data
.done:	mov	edi,edx			; restore edi
	pop	ebx			; pop TOS
	next

; bset ( mem bit -- )
core_header 32,_bset,"bset",flag_none
_bset:
	pop	eax			; eax = base address
	mov	cl,bl
	and	cl,7			; cl = bit modulo 8
	shr	ebx,3			; ebx = bit / 8
	add	ebx,eax			; ebx = ptr to byte we will modify
	mov	ch,1
	shl	ch,cl			; ch = 1 << (bit modulo 8)
	or	[ebx],ch		; set the bit
	pop	ebx			; pop TOS
	next

; bclear ( mem bit -- )
core_header 14,_bclear,"bclear",flag_none
_bclear:
	pop	eax			; eax = base address
	mov	cl,bl
	and	cl,7			; cl = bit modulo 8
	shr	ebx,3			; ebx = bit / 8
	add	ebx,eax			; ebx = ptr to byte we will modify
	mov	ch,1
	shl	ch,cl			; ch = 1 << (bit modulo 8)
	not	ch			; invert ch
	and	[ebx],ch		; clear the bit
	pop	ebx			; pop TOS
	next

; bit ( mem bit -- value )
core_header 1d,_bit,"bit",flag_none
_bit:
	pop	eax			; eax = base address
	mov	cl,bl
	and	cl,7			; cl = bit modulo 8
	shr	ebx,3			; ebx = bit / 8
	add	ebx,eax			; ebx = ptr to byte we will modify
	mov	ch,1
	shl	ch,cl			; ch = 1 << (bit modulo 8)
	and	ch,[ebx]		; set the bit
	jz	.zero
	mov	ebx,-1
	next
.zero:	xor	ebx,ebx
	next

; s= ( a1 a2 -- true|false )
core_header 08,_seq,"s=",flag_none
_seq:
	pop	eax			; eax = a1
	mov	cl,[eax]		; cl = length of a1
	cmp	cl,[ebx]		; compare to length of a1
	jnz	.false			; not equal? return false
	movzx	ecx,cl			; ecx = length of strings
.next:	inc	ebx
	inc	eax
	mov	dl,[ebx]
	cmp	dl,[eax]		; compare a character pair
	jnz	.false
	loop	.next			; end of string?
	mov	ebx,-1			; return true
	next
.false:	xor	ebx,ebx			; return false
	next

; scopy ( src dest -- )
core_header 17,_scopy,"scopy",flag_none
_scopy:
	pop	edx
	movzx	ecx,byte[edx]
	mov	[ebx],cl
	or	ecx,ecx
	jz	.done
.next:	inc	ebx
	inc	edx
	mov	al,[edx]
	mov	[ebx],al
	loop	.next
.done:	pop	ebx
	next

; s>n ( s -- n true | false )
; However, no n is returned if TOS is false on return.
core_header 0b,_ston,"s>n",flag_none
_ston:
	push	edi			; save edi (will be used to store n)
	movzx	ecx,byte[ebx]		; ecx = length of string
	xor	edi,edi			; n = 0
	mov	byte[.negative_flag],0
	cmp	byte[ebx+1],'-'		; negative nr?
	jnz	.not_negative
	mov	byte[.negative_flag],1
	inc	ebx
	dec	ecx
	jz	.error
.not_negative:
.next:	inc	ebx			; go to next character/digit
	mov	al,[ebx]		; al = character
	cmp	al,'0'
	jb	.not_digit
	cmp	al,'9'
	ja	.not_digit		; if(al<'0' || al>'9') goto .not_digit
.digit:	sub	al,'0'			; it is a digit, subtract '0'
.process:
	cmp	al,[_base+5]		; is al >= base of number system?
	jnb	.error			; yes, return error
	movzx	eax,al			; eax = digit value
	push	eax
	mov	eax,[_base+5]		; eax = base of number system
	mul	edi			; multiply with old n
	pop	edx			; edx = digit value
	xchg	eax,edi			; n = base * n + digit
	add	edi,edx
	loop	.next
.done:	xchg	eax,edi			; eax = n
	pop	edi			; restore saved edi
	cmp	byte[.negative_flag],0
	jz	.not_negative_flag
	neg	eax
.not_negative_flag:
	push	eax			; save number
	mov	ebx,-1			; return true
	next
.not_digit:
	cmp	al,'a'
	jb	.error
	cmp	al,'z'
	ja	.error			; if(al<'a' || a>'z') goto .error
.char:	sub	al,'a'-0x0a		; al = digit value
	jmp	short .process
.error:	xor	ebx,ebx			; return false
	pop	edi			; restore saved edi
	next

.negative_flag	db	0

