; mtforth.asm
;
; Glue between multitasking code and the Forth system.

; edi always points to the data area of the process. This is where to store
; process-specific variables.

; remember to free the rstack...

	struc mtforth_struct
.rstack	resd	1			; return stack area
.rsize	resd	1			; return stack size
.data	resd	1			; data area
.code	resd	1			; compiled Forth code
.parent	resd	1			; parent PID
	endstruc


; thread ( datablocksize dstacksize address -- pid )
core_header 26,_thread,"thread",flag_none
_thread:
	pop	edx			; data stack size
	pop	eax			; data block size
	pushf
	cli
	push	esi			; save IP
	push	ebx			; save pointer to code
	push	eax			; save data block size

	mov	ebx,mtforth_entry	; ebx = entry point
	mov	cl,proc_running
	call	proc_create
	; make ecx point to the "data area" in the stack block
	mov	[ebp-4],ecx		; store PID on return stack
	mov	ecx,[eax+proc_struct.stack]
	mov	eax,0x100		; 256b = return stack size
	mov	[ecx+mtforth_struct.rsize],eax
	call	mem_alloc		; esi = return stack mem
	mov	[ecx+mtforth_struct.rstack],esi
	pop	eax
	call	mem_alloc		; esi = data block mem
	mov	[ecx+mtforth_struct.data],esi
	pop	ebx			; pointer to cod
	mov	[ecx+mtforth_struct.code],ebx
	mov	byte[edi+proc_struct.status],proc_running

	pop	esi			; restore IP
	popf
	mov	ebx,[ebp-4]		; return PID
	next

; Here's where execution starts in a new process. Info is stored in the
; start of the stack area (and the stack is of course in the end of it).
mtforth_entry:
	mov	eax,[proc_current]	; eax points to proc_struct
	mov	eax,[eax+proc_struct.stack]
	; initialize return stack
	mov	ebp,[eax+mtforth_struct.rstack]
	add	ebp,[eax+mtforth_struct.rsize]
	mov	edi,[eax+mtforth_struct.data]
	mov	eax,[eax+mtforth_struct.code]
	mov	esi,code_suicide
	jmp	eax
code_suicide:
	dd	mtforth_suicide

core_header 3d,mtforth_suicide,"suicide",flag_none
mtforth_suicide:
	mov	eax,edi
	call	mem_free		; free data area
	mov	esi,[proc_current]
	mov	ecx,[esi+proc_struct.stack]
	mov	eax,[ecx+mtforth_struct.rstack]
	call	mem_free		; free return stack
	call	proc_kill		; esi = proc_current
.die:	jmp	short .die

; parent ( -- parent-PID )
core_header 08,_parent,"parent",flag_none
_parent:
	push	ebx
	mov	ebx,[proc_current]
	mov	ebx,[ebx+proc_struct.parent]
	next

; t@ ( a -- n )
core_header 12,_tfetch,"t@",flag_none
_tfetch:
	mov	ebx,[edi+ebx]
	next

; t! ( n a -- )
core_header 14,_tstore,"t!",flag_none
_tstore:
	pop	eax
	mov	[edi+ebx],eax
	pop	ebx
	next

; tc@ ( a -- n )
core_header 19,_tcfetch,"tc@",flag_none
_tcfetch:
	movzx	ebx,byte[edi+ebx]
	next

; tc! ( n a -- )
core_header 19,_tcstore,"tc!",flag_none
_tcstore:
	pop	eax
	mov	[edi+ebx],al
	pop	ebx
	next

