; f.asm
;
; Entry file for the F Forth environment.

[bits 32]
[section .text]

%define memory_size	0x1000		; 1kB blocks > 1MB
%define mem_table_size	0x1000		; 4096 items (about 40kB)
%define my_stack_size	0x1000		; in bytes
%define rstack_size	0x400		; in bytes
%define pit_stack_size	0x400		; won't need much more than 1kB

%define registers_size	152

start:
	call	gdt_init
	call	mem_init
	call	idt_init
	call	irq_init
	call	kbd_init
	call	crt_init
	call	pit_init

	mov	eax,my_stack_size
	call	mem_alloc
	mov	esp,esi
	add	esp,my_stack_size-4
	mov	edx,esi
	call	proc_init		; enter multitasking mode

	mov	eax,rstack_size
	call	mem_alloc

	mov	eax,[proc_current]
	mov	eax,[eax+proc_struct.stack]
	mov	dword[eax+mtforth_struct.code],0
	mov	dword[eax+mtforth_struct.data],0
	mov	dword[eax+mtforth_struct.rstack],esi
	add	esi,rstack_size
	mov	ebp,esi
	mov	dword[eax+mtforth_struct.rsize],rstack_size

	sti				; do not forget :)

	mov	ebx,main_screen
	mov	esi,str_test
	call	crt_put_string

mainword:
	call	_nest
	dd	_dolit,'A',_emit
	dd	_dolit,0x100,_dolit,0x100,_dolit,testword
	dd	_thread
.loop:
	dd	_dolit,recvdata
	dd	_ipc_recv,_drop
	dd	_dolit,recvdata,_cfetch,_emit
	dd	_branch,.loop

recvdata	dd	0,0,0,0
senddata	dd	0,0,0,0

testword:
	call	_nest
	dd	_dolit,'X',_dolit,senddata,_cstore,_dolit,senddata
	dd	_parent,_ipc_send,_dolit,'P',_emit,_delay
	dd	_dolit,'Y',_dolit,senddata,_cstore,_dolit,senddata
	dd	_dolit,1,_ipc_send,_dolit,'Q',_emit,_delay
	dd	_dolit,'Z',_dolit,senddata,_cstore,_dolit,senddata
	dd	_dolit,1,_ipc_send,_dolit,'R',_emit,_delay
	dd	_exit

test_proc:
	xor	ecx,ecx
	mov	eax,100
	call	timer_delay
	mov	dl,'Q'
	mov	ebx,main_screen
	call	crt_put_char
	jmp	test_proc

str_test	db	0x0e,"Welcome to F!",0x0a

%include "gdt.asm"
%include "idt.asm"
%include "irq.asm"
%include "mm.asm"
%include "kbd.asm"
%include "crt.asm"
%include "pit.asm"
%include "timer.asm"
%include "proc.asm"
%include "ipc.asm"

%include "coreheader.asm"

%include "core.asm"
%include "cpu.asm"
%include "stack.asm"
%include "math.asm"
%include "mem.asm"
%include "jump.asm"
%include "dict.asm"
%include "mtforth.asm"
%include "ipcforth.asm"
%include "compiler.asm"

core_header 32,_emit,"emit",flag_none
_emit:
	mov	dl,bl
	mov	ebx,main_screen
	call	crt_put_char
	pop	ebx
	next

_delay:
	mov	eax,100
	xor	ecx,ecx
	call	timer_delay
	next

; This file has to be included after all word definitions.

%include "coredict.asm"

[section .bss]

proc_registers	resb	registers_size
pit_stack	resb	pit_stack_size
mem_table	resb	mem_table_size*9
kbd_buffer	resb	kbd_buffer_size
main_screen	resb	screen_size
idt		resb	0x800

heap:

