; mm.asm
;
; Memory manager, Second Edition(tm)
;
; The first one was stolen from Primula, and it's really, really ugly.
; So now when I'm sitting here alone, with no Internet connection, I
; might as well write a new one.

%define mem_status_free	0
%define mem_status_used	1

	struc mem_header
.next	resd	1			; pointer to next header
.last	resd	1			; pointer to last header
.size	resd	1			; size of block in bytes
.status	resb	1			; free/used?
	endstruc

; eax = memory to allocate
; Returns:
; esi = memory block (or 0 on error)

mem_init:
	mov	dword[0x100000+mem_header.next],0
	mov	dword[0x100000+mem_header.last],0
mov dword[0x100000+mem_header.size],(memory_size * 0x400)-mem_header_size
	mov	byte[0x100000+mem_header.status],mem_status_free
	ret

mem_alloc:
	pushf
	cli				; we do NOT want to be disturbed here
	push	ecx
	push	edx
	push	ebx
	or	eax,eax
	jz	.error			; block size = 0? don't allocate
	add	eax,byte mem_header_size

	mov	ecx,[mem_first]		; ecx points to memory block
.try_block:
	cmp	byte[ecx+mem_header.status],mem_status_free
	jnz	.next_block		; make sure the block is free
	cmp	dword[ecx+mem_header.size],eax
	jb	.next_block		; and also large enough
	jnz	.resize			; if it's not an exact match, resize

	mov	byte[ecx+mem_header.status],mem_status_used
	lea	esi,[ecx+mem_header_size]
	jmp	short .quit
.resize:
	lea	esi,[ecx+mem_header_size]
	; edx = address of new header
	lea	edx,[ecx+eax]
	mov	byte[ecx+mem_header.status],mem_status_used
	sub	eax,byte mem_header_size
	; store size of allocated block
	xchg	eax,[ecx+mem_header.size]
	; save pointer to next block
	mov	ebx,[ecx+mem_header.next]
	; make this one point to the new block
	mov	[ecx+mem_header.next],edx
	; and the new block to the block after
	mov	[edx+mem_header.next],ebx
	mov	[edx+mem_header.last],ecx
	or	ebx,ebx
	jz	.no_next
	mov	[ebx+mem_header.last],edx
.no_next:
	sub	eax,[ecx+mem_header.size]
	sub	eax,byte mem_header_size ; eax = size of created block
	mov	[edx+mem_header.size],eax
	mov	byte[edx+mem_header.status],mem_status_free
	jmp	short .quit
.next_block:
	mov	ecx,[ecx+mem_header.next]
	or	ecx,ecx			; last block?
	jnz	.try_block		; yes, return NULL
.error:
	xor	esi,esi
.quit:
	pop	ebx
	pop	edx
	pop	ecx
	popf
	ret

; eax = memory block to free
mem_free:
	pushf
	cli
	push	ecx
	push	edx
	or	eax,eax
	jz	.quit

	sub	eax,byte mem_header_size ; eax = mem_header struct of block
	mov	byte[eax+mem_header.status],mem_status_free
	mov	ecx,[eax+mem_header.next]
	or	ecx,ecx
	jz	.no_next
	cmp	byte[ecx+mem_header.status],mem_status_free
	jnz	.no_next
	mov	edx,[ecx+mem_header.size]
	mov	ecx,[ecx+mem_header.next]
	add	edx,byte mem_header_size
	add	[eax+mem_header.size],edx
	mov	[eax+mem_header.next],ecx
.no_next:
	mov	ecx,[eax+mem_header.last]
	or	ecx,ecx
	jz	.no_last
	cmp	byte[ecx+mem_header.status],mem_status_free
	jnz	.no_last
	mov	edx,[eax+mem_header.size]
	mov	eax,[eax+mem_header.next]
	add	edx,byte mem_header_size
	add	[ecx+mem_header.size],edx
	mov	[ecx+mem_header.next],eax
.no_last:
.quit:
	pop	edx
	pop	ecx
	xor	eax,eax
	popf
	ret


mem_first:
	dd	0x100000		; 1MB

