; pit.asm
;
; Part of the multitasking & timer systems.
;

; pit_interrupt
;
; this is the interrupt handler connected to int 0x20 (IRQ0)

pit_interrupt:
	cld
	fsave	[proc_registers+44]	; save FPU status
	mov	dword[proc_registers+00],eax	; save all registers
	mov	dword[proc_registers+04],ecx
	mov	dword[proc_registers+08],edx
	mov	dword[proc_registers+12],ebx
	mov	dword[proc_registers+16],esp
	mov	dword[proc_registers+20],ebp
	mov	dword[proc_registers+24],esi
	mov	dword[proc_registers+28],edi
	mov	ecx,esp			; ecx points to stack
	mov	eax,dword[ecx]		; eax = return eip
	mov	dword[proc_registers+32],eax
	mov	eax,dword[ecx+4]	; eax = return cs
	mov	dword[proc_registers+36],eax
	mov	eax,dword[ecx+8]	; eax = saved eflags
	mov	dword[proc_registers+40],eax
					; use our internal stack
	mov	esp,pit_stack+pit_stack_size-4

	cmp	dword[fdc_timeout],0
	jz	.timed_out
	dec	dword[fdc_timeout]
	jnz	.timed_out
	mov	dx,0x3f2
	mov	al,0x0c
	out	dx,al			; turn off floppy drive motor
.timed_out:
	call	timer_handler
	mov	al,0x20			; tell PIC we're done
	out	0x20,al
	frstor	[proc_registers+44]	; load saved FPU statuss
	mov	ecx,dword[proc_registers+16]	; ecx = saved esp
	mov	eax,dword[proc_registers+32]	; eax = saved return eip
	mov	dword[ecx],eax
	mov	eax,dword[proc_registers+36]	; eax = saved return cs
	mov	dword[ecx+4],eax
	mov	eax,dword[proc_registers+40]	; eax = saved eflags
	mov	dword[ecx+8],eax
	mov	eax,dword[proc_registers+00]	; load registers
	mov	ecx,dword[proc_registers+04]
	mov	edx,dword[proc_registers+08]
	mov	ebx,dword[proc_registers+12]
	mov	esp,dword[proc_registers+16]
	mov	ebp,dword[proc_registers+20]
	mov	esi,dword[proc_registers+24]
	mov	edi,dword[proc_registers+28]
	iret

; pit_init
;
; all register unchanged

pit_init:
	push	eax
	push	ecx
	push	ebx

	mov	cx,1000
	call	pit_set_freq		; set PIT frequency (1000Hz)

	mov	eax,0x20		; set interrupt vector 0x20
	mov	ebx,pit_interrupt	; our interrupt handler
	call	idt_set_int

	mov	cl,0			; let's enable IRQ0
	call	irq_enable

	pop	ebx
	pop	ecx
	pop	eax
	ret

; pit_set_freq
;
; cx = PIT frequency in Hz

pit_set_freq:
	push	eax
	push	edx

	mov	al,0x34
	out	0x43,al			; set frequency

	; now we should divide 0x1234dd with the parameter
	; set dx:ax to 0x1234dd, and cx to the parameter
	mov	dx,0x0012
	mov	ax,0x34dd
	; divide 0x1234dd with the parameter, place the result in ax
	div	cx
	out	0x40,al			; LSB already in al
	mov	al,ah
	out	0x40,al			; MSB

	pop	edx
	pop	eax
	ret

