; proc.asm
;
; Processes.
;
; 2003-10-29:
; Bug fixed in proc_create.. all this time esp has been pointing in the wrong
; direction. Grrr!

proc_running	equ	0
proc_sleeping	equ	1

	struc proc_struct
.registers	resb	registers_size	; size of all the registers (152)
.next		resd	1		; pointer to next proc_struct
.prev		resd	1		; pointer to previous proc_struct
.stack		resd	1		; points to the stack, has to be freed!
; .info		resd	1		; pointer to process info struct
.status		resb	1		; PROC_RUNNING / PROC_SLEEPING
.nosave		resb	1		; should we save registers?
.id		resd	1		; PID
.ipc		resd	1		; ipc_list pointer.
.parent		resd	1		; parent PID
	endstruc

; proc_next
;
; Interrupt function to switch to next process.

proc_next:
	fsave	[proc_registers+44]	; save FPU status
	mov	dword[proc_registers+00],eax	; save all registers
	mov	dword[proc_registers+04],ecx
	mov	dword[proc_registers+08],edx
	mov	dword[proc_registers+12],ebx
	mov	dword[proc_registers+16],esp
	mov	dword[proc_registers+20],ebp
	mov	dword[proc_registers+24],esi
	mov	dword[proc_registers+28],edi
	mov	ecx,esp			; ecx points to stack
	mov	eax,dword[ecx]		; eax = return eip
	mov	dword[proc_registers+32],eax
	mov	eax,dword[ecx+4]	; eax = return cs
	mov	dword[proc_registers+36],eax
	mov	eax,dword[ecx+8]	; eax = saved eflags
	mov	dword[proc_registers+40],eax
					; use our internal stack
	mov	esp,pit_stack+pit_stack_size-4
	call	proc_switch
	frstor	[proc_registers+44]	; load saved FPU statuss
	mov	ecx,dword[proc_registers+16]	; ecx = saved esp
	mov	eax,dword[proc_registers+32]	; eax = saved return eip
	mov	dword[ecx],eax
	mov	eax,dword[proc_registers+36]	; eax = saved return cs
	mov	dword[ecx+4],eax
	mov	eax,dword[proc_registers+40]	; eax = saved eflags
	mov	dword[ecx+8],eax
	mov	eax,dword[proc_registers+00]	; load registers
	mov	ecx,dword[proc_registers+04]
	mov	edx,dword[proc_registers+08]
	mov	ebx,dword[proc_registers+12]
	mov	esp,dword[proc_registers+16]
	mov	ebp,dword[proc_registers+20]
	mov	esi,dword[proc_registers+24]
	mov	edi,dword[proc_registers+28]
	iret

; proc_switch
;
; switch process
;
; returns:
; all registers unchanged

proc_switch:
	push	eax
	push	ecx
	push	esi
	push	edi

	mov	eax,dword[proc_current]
	cmp	byte[eax+proc_struct.nosave],0
	jnz	.dont_save		; don't save registers...
	lea	edi,[eax+proc_struct.registers]
	mov	esi,proc_registers	; copy from register buffer
					; to the processes registers
	mov	ecx,registers_size / 4
	rep	movsd
.dont_save:
	mov	byte[eax+proc_struct.nosave],0
.try_next:
	mov	eax,dword[eax+proc_struct.next]
	cmp	byte[eax+proc_struct.status],proc_running
	jnz	.try_next
	mov	dword[proc_current],eax
	lea	esi,[eax+proc_struct.registers]
	mov	edi,proc_registers	; copy from the processes registers
					; to the buffer
	mov	ecx,registers_size / 4
	rep	movsd

	pop	edi
	pop	esi
	pop	ecx
	pop	eax
	ret

; proc_init
;
; enters multitasking mode, and creates the first thread
;
; edx = stack of current process (should be able to be deallocated)
;
; returns:
; all registers unchanged

proc_init:
	push	eax
	push	esi
	push	edi

	mov	eax,proc_struct_size
	call	mem_alloc		; allocate sizeof(proc_struct)

					; bytes memory, pointer in esi

	mov	dword[proc_current],esi
	; now initialize the first process before we enter multitasking mode
	mov	dword[esi+proc_struct.next],esi
	mov	dword[esi+proc_struct.prev],esi
	mov	byte[esi+proc_struct.status],proc_running
	mov	byte[esi+proc_struct.nosave],0
	mov	dword[esi+proc_struct.stack],edx
	mov	dword[esi+proc_struct.ipc],0

	mov	eax,dword[proc_current_id]
	inc	dword[proc_current_id]
	mov	dword[esi+proc_struct.id],eax
	mov	byte[proc_mt],1		; enter multitasking mode

	mov	eax,0x30		; set int vector 0x30 (switch task)
	mov	ebx,proc_next
	call	idt_set_int

	pop	edi
	pop	esi
	pop	eax
	ret

; proc_create
;
; creates a new process
;
; edx = stack size
; ebx = starting point
; cl = status (PROC_RUNNING, 0, or PROC_SLEEPING, 1)
;
; returns:
; eax = pointer to proc struct, 0 if error
; ecx = PID, 0 if error

proc_create:
	pushf
	cli
	push	esi
	push	edi

	mov	eax,proc_struct_size
	call	mem_alloc		; allocate memory for the proc_struct

	or	esi,esi
	jnz	.mem_allocated		; memory not allocated, return error
	xor	eax,eax
	jmp	.quit
.mem_allocated:
	mov	byte[esi+proc_struct.status],cl
	mov	byte[esi+proc_struct.nosave],0

	xor	eax,eax
	lea	edi,[esi+proc_struct.registers]
	mov	ecx,registers_size / 4
	rep	stosd			; empty the registers fiels

	mov	edi,esi

	mov	eax,edx			; eax = stack size to allocate
	call	mem_alloc		; try to allocate it
	or	esi,esi			; did we fail?
	jnz	.stack_allocated	; yeah, return 0

	xor	eax,eax
	mov	ecx,eax
	jmp	 .quit
.stack_allocated:
	mov	dword[edi+proc_struct.stack],esi

;	shr	edx,2
;	sub	edx,byte 16
; What did I smoke when I wrote this?!

	sub	edx,byte 4
	add	esi,edx			; set esp
	mov	dword[edi+proc_struct.registers+16],esi
	mov	esi,edi			; now we set the cs:eip and eflags
	mov	dword[esi+proc_struct.registers+32],ebx
	mov	dword[esi+proc_struct.registers+36],8
	mov	dword[esi+proc_struct.registers+40],0x00000202

	mov	dword[esi+proc_struct.ipc],0
	mov	eax,[proc_current]
	mov	eax,[eax+proc_struct.id]
	mov	dword[esi+proc_struct.parent],eax

	mov	eax,dword[proc_current_id]
	inc	dword[proc_current_id]
	mov	dword[esi+proc_struct.id],eax

	mov	eax,dword[proc_current]

	; process->prev = proc_current;
	mov	dword[esi+proc_struct.prev],eax

	; process->next = proc_current->next;
	mov	ecx,dword[eax+proc_struct.next]
	mov	dword[esi+proc_struct.next],ecx

	; proc_current->next->prev = process;
	mov	dword[ecx+proc_struct.prev],esi

	; proc_current->next = process;
	mov	dword[eax+proc_struct.next],esi

	mov	eax,esi			; return ptr to out proc struct in eax
	mov	ecx,dword[esi+proc_struct.id]	; return PID in ecx
.quit:
	pop	edi
	pop	esi
	popf
	ret

; proc_kill
;
; kills a process
;
; esi = pointer to proc_struct
;
; returns:
; all registers unchanged

proc_kill:
	push	eax
	push	ecx
	push	edx

	xor	edx,edx			; suicide flag off
	cmp	esi,dword[proc_current]
	jnz	.not_suicide

	mov	dl,1			; suicide flag on, how cute...
.not_suicide:
	mov	eax,dword[esi+proc_struct.stack]
	call	mem_free		; free the processes stack

	; process->next->prev = process->prev
	mov	eax,dword[esi+proc_struct.next]
	mov	ecx,dword[esi+proc_struct.prev]
	mov	dword[eax+proc_struct.prev],ecx

	; process->prev->next = process->next
	mov	eax,dword[esi+proc_struct.next]
	mov	dword[ecx+proc_struct.next],eax

	or	dl,dl			; suicide?
	jnz	.nope			; y, set proc_current to process->next
	mov	eax,dword[esi+proc_struct.next]
	mov	dword[proc_current],eax
.nope:
	or	dl,dl			; suicide?
	jz	.done			; nope, skip this
	mov	eax,dword[proc_current] ; proc_current = proc_current->next ?
	cmp	eax,dword[eax+proc_struct.next]
	jnz	.done			; no, skip this

	; now set 'nosave' to 1
	mov	byte[eax+proc_struct.nosave],1
.done:
	xchg	eax,esi			; eax points to our process
	call	mem_free
.quit:
	pop	edx
	pop	ecx
	pop	eax
	ret

; proc_get_pointer
;
; Gets the pointer to a proc structure from its PID.
;
; eax = PID
;
; returns:
; eax = pointer, 0 if error

proc_get_pointer:
	push	ecx
	push	edx

	mov	ecx,dword[proc_current]	; start at this element in the list
	mov	edx,ecx			; save a pointer to this, so we can
					; stop when we reach it again
.again:
	; compare ID with the one we're looking for
	cmp	dword[ecx+proc_struct.id],eax
	jz	.found
	mov	ecx,dword[ecx+proc_struct.next]
	; have we reached the element we started at?
	cmp	ecx,edx			
	jnz	.again			; nope, try again
	xor	eax,eax
	jmp	short .quit
.found:
	xchg	eax,ecx			; eax = pointer to struct
.quit:
	pop	edx
	pop	ecx
	ret

; proc_get_current_pid
;
; returns:
; eax = PID

proc_get_current_pid:
	mov	eax,dword[proc_current]
	mov	eax,dword[eax+proc_struct.id]
	ret

proc_mt		db	0		; multitasking is off by default
proc_current_id	dd	1		; first PID is 1
proc_current	dd	0

