#ifndef __block_h__
#define __block_h__

/*
 * Disk structure
 *
 *  boot block
 *  block allocation table (bit map)
 *  root block
 *  kernel
 *  main block describing the core source blocks
 *  core source blocks
 */

typedef struct {
	uint8_t jump[2];		/* jump code */
	uint8_t heads;			/* number of heads */
	uint16_t tph;			/* tracks per head */
	uint8_t spt;			/* sectors per track */
	uint32_t blocks;		/* disk size in blocks */
	uint32_t base;			/* base block */
	uint32_t kernel;		/* start of kernel */
	uint8_t kernelsize;		/* size of kernel, in blocks */
	uint8_t code[491];		/* boot code */
	uint16_t id;			/* 0xaa55 */
	uint8_t fill[512];		/* zeroes */
} __attribute__((packed)) bootblock;

typedef struct {
	uint32_t block;			/* the block we are refering to */
	uint8_t flags;			/* bit 1 set if the block is a main block */
	uint8_t namelen;		/* length of block name */
	uint8_t name[10];		/* block name */
} __attribute__((packed)) subblock;

typedef struct {
	subblock sub[0x40];
}  __attribute__((packed)) mainblock;

#endif

