#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include "block.h"

void setbit(uint8_t *a, int bit) {
	a[bit>>3] |= 1<<(bit&7);
}

int main(int argc, char **argv) {
	FILE *dev;
	FILE *boot;
	FILE *kernel;
	FILE *core;
	bootblock bootbuf;
	mainblock rootbuf,mainbuf;
	int kernelbytes,corebytes;
	uint8_t *kernelbuf;
	uint8_t *corebuf;
	uint8_t bat[0x400];
	int batsize;
	int kernelsize,coresize,coreblock;
	int i;

	if(argc != 5) {
		printf("usage: mkdist device bootloader kernel core\n");
		return 1;
	}

	printf("sizeof(bootblock) = %d, sizeof(mainblock) = %d\n",
			sizeof(bootblock),sizeof(mainblock));

	if((dev = fopen(argv[1],"r+")) == NULL) {
		printf("Can not open device file.\n");
		return 1;
	}

	if((boot = fopen(argv[2],"r")) == NULL) {
		printf("Can not open boot loader.\n");
		return 1;
	}
	if(fread(&bootbuf,sizeof(bootblock),1,boot) != 1) {
		printf("Can not read 1kB from boot loader file.\n");
		return 1;
	}
	fclose(boot);

	if((kernel = fopen(argv[3],"r")) == NULL) {
		printf("Can not open kernel binary.\n");
		return 1;
	}
	fseek(kernel,0,SEEK_END);
	kernelbytes = ftell(kernel);
	kernelsize = ((kernelbytes-1)>>10)+1;
	kernelbuf = malloc(kernelsize<<10);
	fseek(kernel,0,SEEK_SET);
	fread(kernelbuf,kernelbytes,1,kernel);
	fclose(kernel);

	if((core = fopen(argv[4],"r")) == NULL) {
		printf("Can not open core system sources.\n");
		return 1;
	}
	fseek(core,0,SEEK_END);
	corebytes = ftell(core)+1;
	coresize = ((corebytes-1)>>10)+1;
	corebuf = malloc(coresize<<10);
	fseek(core,0,SEEK_SET);
	fread(corebuf,corebytes-1,1,core);
	corebuf[corebytes-1] = 0;
	fclose(core);

	bootbuf.heads = 2;
	bootbuf.tph = 0x50;
	bootbuf.spt = 0x12;
	bootbuf.blocks = 1440;
	bootbuf.base = 0;
	bootbuf.kernel = 3;
	bootbuf.kernelsize = kernelsize;
	bootbuf.id = 0xaa55;

	batsize = 1;
	memset(bat,0,0x400);
	setbit(bat,0);		/* boot block */
	setbit(bat,1);		/* boot allocatoin table */
	setbit(bat,2);		/* root block */
	for(i=bootbuf.kernel;i<bootbuf.kernel+bootbuf.kernelsize;i++) {
		setbit(bat,i);	/* kernel */
	}
	coreblock = bootbuf.kernel+kernelsize+1;
	setbit(bat,bootbuf.kernel+kernelsize);
	for(i=coreblock;i<coreblock+coresize;i++) {
		setbit(bat,i);
	}

	memset(&rootbuf,0,0x400);
	memset(&mainbuf,0,0x400);

	rootbuf.sub[0].block = bootbuf.kernel + kernelsize;
	rootbuf.sub[0].flags = 1;
	rootbuf.sub[0].namelen = 4;
	strcpy(rootbuf.sub[0].name,"core");
	rootbuf.sub[1].block = 0xffffffff;

	for(i=0;i<coresize;i++) {
		mainbuf.sub[i].block = coreblock + i;
		mainbuf.sub[i].flags = 0;
		mainbuf.sub[i].namelen = 0;
	}
	mainbuf.sub[coresize].block = 0xffffffff;

	printf(	"Preparing to write...\n\n"
			"Block allocation table size: %d\n"
			"Kernel size:                 %d\n"
			"Core language sources size:  %d\n",
			(int)batsize,(int)kernelsize,(int)coresize);

	fwrite(&bootbuf,0x400,1,dev);
	fwrite(bat,0x400,batsize,dev);
	fwrite(&rootbuf,0x400,1,dev);
	fwrite(kernelbuf,0x400,kernelsize,dev);
	fwrite(&mainbuf,0x400,1,dev);
	fwrite(corebuf,0x400,coresize,dev);

	printf("OK.\n");

	fclose(dev);
	free(corebuf);
	free(kernelbuf);
	return 0;
}

